import abc

class AbstractClassifier(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __init__(self, classes):
        self.classes = {clas:idx for idx,clas in enumerate(classes)}
        self.revclasses = {idx:clas for idx,clas in enumerate(classes)}

    def x_to_modelspace(self, x): return x
    def y_to_modelspace(self, y): return y

    @abc.abstractmethod
    def score(self, x, classes=None):
        pass
    def score_on_batch(self, batch, classes=None):
        if classes is None:
            classes = [None] * len(batch)
        return [self.score(x, classes) for x, classes in zip(batch, classes)]

    def predict(self, x, classes=None, argmax=True, scores=False, raw=False):
        raw_scores = self.score(x, classes)
        if classes is None:
            classes = self.classes
        else:
            raw_scores = [s*int(self.revclasses[i] in classes) for i,s in enumerate(raw_scores)]
        argmax = self.revclasses[max(enumerate(raw_scores), key=lambda x:x[1])[0]] if argmax else None
        scores =  {c: raw_scores[self.classes[c]] for c in classes} if scores else None
        raw = raw_scores if raw else None
        out = tuple(x for x in (argmax, scores, raw) if x is not None)
        return out[0] if len(out) == 1 else out
    @abc.abstractmethod
    def update(self, y, _y, _x, x=None, extra=None):
        pass

    def predict_on_batch(self, batch, classes=None, argmax=True, scores=False, raw=False):
        if classes is None:
            classes = [None] * len(batch)
        return [self.predict(x, classes, argmax, scores, raw) for x, classes in zip(batch, classes)]
    def train_on_batch(self, *batch):
        return [self.update(*x) for x in zip(*batch)]


    def dump_features(self, filehandler):
        import pickle
        if filehandler is not None:
            try:
                _ = self.real_score
            except AttributeError:
                self.real_score = self.score
            def score_with_log(x, c=None):
                if hasattr(filehandler, 'lock_dump_features'):
                    filehandler.lock_dump_features.acquire()
                pickle.dump(x, filehandler)
                if hasattr(filehandler, 'lock_dump_features'):
                    filehandler.lock_dump_features.release()
                return self.real_score(x, c)
            self.score = score_with_log
        else:
            try:
                self.score = self.real_score
            except AttributeError:
                pass

    def polish(self):
        pass

    def backup(self):
        pass

    def restore(self, state):
        pass

    def with_polished(self, measure):
        state = self.backup()
        self.polish()
        ret = measure()
        self.restore(state)
        return ret
