from .AbstractClassifier import AbstractClassifier

class Perceptron(AbstractClassifier):
    gradient_normalization = 1

    def __init__(self, classes):
        super().__init__(classes)
        # Each feature gets its own weight vector, so weights is a dict-of-
        # arrays
        self.weights = {}
        # The accumulated values and the last time the feature was changed,
        # for the averaging. Also dict-of-arrays, keyed by features
        self._cumul = {}
        # Number of instances seen
        self.i = 0

    def score(self, features, classes=None):
        classes = [(clas,self.classes[clas]) for clas in classes] if classes is not None else self.classes.items()
        class_ids = [idx for _,idx in classes]
        scores = [0 for _ in self.classes]
        all_weights = self.weights
        for feat, value in features.items():
            if value == 0:
                continue
            if feat not in all_weights:
                continue
            weights = all_weights[feat]
            for idx in class_ids:
                scores[idx] += value * weights[idx]
        return scores

    def update(self, y, _y, _x, x=None, extra=None):
        if not extra or extra.get('time_step', True):
            self.i += 1
        if y == _y and (_x is None or x == _x):
            return

        def upd_feat(c, f, w, v):
            total, tstamp = self._cumul[f][c]
            self._cumul[f][c] = (total + (self.i - tstamp) * w, self.i)
            self.weights[f][c] = w + v
        def upd_feats(c, fs, alpha):
            for f in fs:
                weights = self.weights.get(f)
                if weights is None:
                    weights = [0.0 for _ in self.classes]
                    self.weights[f] = weights
                if f not in self._cumul:
                    self._cumul[f] = [(0,0) for _ in self.classes]
                upd_feat(c, f, weights[c], alpha * fs[f])

        alpha = (extra.get('alpha', 1.0) if extra else 1.0) / self.gradient_normalization
        if x is None: x = _x
        # separate '+' and '-' updates
        upd_feats(self.classes[_y], _x, -alpha)
        upd_feats(self.classes[y], x, alpha)


    def average_weights(self):
        for feat, weights in self.weights.items():
            new_feat_weights = [0 for _ in weights]
            if feat not in self._cumul:
                self._cumul[feat] = [(0,0) for _ in self.classes]
            for clas, (weight, (total, tstamp)) in enumerate(zip(weights, self._cumul[feat])):
                total += (self.i - tstamp) * weight
                new_feat_weights[clas] = total / float(self.i)
            self.weights[feat] = new_feat_weights
        self._cumul = {}

    def full_backup(self):
        from copy import deepcopy
        return deepcopy((self.weights, self._cumul, self.i))

    def backup(self):
        from copy import copy
        return copy(self.weights), self._cumul, self.i

    def restore(self, state):
        self.weights, self._cumul, self.i = state

    def polish(self):
        self.average_weights()
