from .AbstractClassifier import AbstractClassifier

class JointClassifier(AbstractClassifier):
    def __init__(self, *parts):
        self.parts = parts

    def x_to_modelspace(self, x):
        i, x = x
        return i, self.parts[i].x_to_modelspace(x)
    def y_to_modelspace(self, y):
        i, y = y
        return i, self.parts[i].y_to_modelspace(y)

    def score(self, x, classes=(0,None)):
        return self.parts[x[0]].score(x[1], classes[1])
    def score_on_batch(self, batch, classes=None):
        fs = [part.score_on_batch for part in self.parts]
        return apply_on_batch_parts(fs, *batch, classes=classes) #

    def predict(self, x, classes=None, argmax=True, scores=False, raw=False):
        return self.parts[x[0]].predict(x[1], classes, argmax, scores, raw)
    def update(self, y, _y, _x, x=(0,None), extra=None):
        self.parts[y[0]].update(y[1], _y[1], _x[1], x[1], extra)

    def predict_on_batch(self, batch, classes=None, argmax=True, scores=False, raw=False):
        fs = [part.predict_on_batch for part in self.parts]
        return apply_on_batch_parts(fs, batch, classes, argmax=argmax, scores=scores, raw=raw) #
    def train_on_batch(self, *batch):
        fs = [part.train_on_batch for part in self.parts]
        return apply_on_batch_parts(fs, *batch, indexed=4)

    def dump_features(self, filehandler):
        import pickle
        for i, model in enumerate(self.parts):
            if filehandler is not None:
                try:
                    _ = model.real_score
                except AttributeError:
                    model.real_score = model.score
                def score_with_log(x, c=None):
                    if hasattr(filehandler, 'lock_dump_features'):
                        filehandler.lock_dump_features.acquire()
                    pickle.dump((i,x), filehandler)
                    if hasattr(filehandler, 'lock_dump_features'):
                        filehandler.lock_dump_features.release()
                    return self.real_score(x, c)
                model.score = score_with_log
            else:
                try:
                    model.score = model.real_score
                except AttributeError:
                    pass

    def polish(self):
        for model in self.parts:
            model.polish()

    def backup(self):
        return tuple(model.backup() for model in self.parts)

    def restore(self, state):
        for model, state in zip(self.parts, state):
            model.restore(state)

def apply_on_batch_parts(fs, *batch, indexed=1, **kwargs):
    from misc.itertools import deinterleave
    # default values for kwargs
    batch = [arg if type(arg) == list else [arg] * len(batch[0]) for arg in batch]
    # pad non-indexed batches with selector=None
    batch = batch[:indexed] + [[(None,v) for v in arg] for arg in batch[indexed:]]
    # retrieve tuples, with selector in front
    batch = [(tup[0][0], [arg[1] for arg in tup]) for tup in zip(*batch)]
    # add batch-position markers on tuples
    batch = [(s, (pos, tup)) for pos, (s, tup) in enumerate(batch)]
    # split according to selectors
    batches = deinterleave(batch, len(fs))
    # extract position markers from tuples
    batches = [tuple(zip(*batch)) for batch in batches]
    batches = [batch if batch else ((),()) for batch in batches]
    positions, batches = zip(*batches)
    # back to a tuple of batches
    batches = [list(map(list,zip(*batch))) for batch in batches]
    # call each f
    applied = [f(*batch, **kwargs) if batch else None for f, batch in zip(fs,batches)]
    # restore output lists, positions and order
    applied = [(positions, outputs if outputs else tuple(None for _ in positions)) for positions, outputs in zip(positions, applied)]
    outputs = [x for positions, outputs in applied for x in zip(positions, outputs)]
    return [out for _,out in sorted(outputs)]
