from .AbstractClassifier import AbstractClassifier

import abc

import numpy as np
from keras.utils import np_utils

def weighted_linear_loss(gold, pY):
    from keras import backend as K
    alpha, Y = gold[:,0], gold[:,1:]
    loss = - alpha * K.sum(pY*Y, axis=1)
    return K.sum(loss)

class NeuralClassifier(AbstractClassifier):
    gradient_normalization = 1

    @abc.abstractmethod
    def __init__(self, classes, voc):
        super().__init__(classes)
        def indexes(voc): return {v:i for i, v in enumerate(voc, start=2)}
        if type(voc) == tuple:
            self.voc = tuple(indexes(v) for v in voc)
        elif type(voc) == dict:
            self.voc = {k: indexes(v) for k, v in voc.items()}
        else:
            self.voc = indexes(voc)

    def x_to_modelspace(self, x):
        if type(x) == tuple:
            return [np.array([voc.get(e,1) for e in subx]) for voc,subx in zip(self.voc,x)]
        if type(x) == dict:
            return {k:np.array([self.voc[k].get(e,1) for e in subx]) for k,subx in x.items()}
        return np.array([self.voc.get(e,1) for e in x])
    def y_to_modelspace(self, y):
        y = self.classes[y]
        y = np_utils.to_categorical(self.batch2np([y]), len(self.classes))[0]
        return y

    def batch2np(self, batch):
        batch = list(batch)
        if batch and type(batch[0]) == list: # x as multi-inputs
            return [np.array(subbatch) for subbatch in zip(*batch)]
        if batch and type(batch[0]) == dict: # x as named multi-inputs
            return {k:np.array([x[k] for x in batch]) for k in batch[0].keys()}
        if batch and type(batch[0]) == tuple: # y with extra data
            return np.array([np.array(y, dtype=object) for y in batch])
        return np.array(batch)

    def score(self, x, classes=None):
        return self.score_on_batch([x])[0]
    def score_on_batch(self, xs, classes=None):
        xs = self.batch2np(xs)
        return self.model.predict_on_batch(xs)

    def predict(self, x, classes=None, argmax=True, scores=False, raw=False):
        return self.predict_on_batch([x], [classes], argmax, scores, raw)[0]

    def update(self, y, _y, _x, x=None, extra=None):
        return self.train_on_batch([y], [_y], [_x], [x], [extra])

    def predict_on_batch(self, _xs, classes=None, argmax=True, scores=False, raw=False):
        if not _xs: return _xs
        raw_scores = self.score_on_batch(_xs)
        if classes is None:
            classes = [self.classes] * len(_xs)
        raw_scores = np.array([scores if classes is None else scores*np.array([self.revclasses[i] in classes for i in range(len(scores))]) for scores, classes in zip(raw_scores, classes)])
        argmax = [self.revclasses[y] for y in np_utils.probas_to_classes(raw_scores)] if argmax else None
        scores = [{c: scores[self.classes[c]] for c in (self.classes if classes is None else classes)} for scores, classes in zip(raw_scores, classes)] if scores else None
        raw = raw_scores if raw else None
        out = tuple(x for x in (argmax, scores, raw) if x is not None)
        if len(out) == 1:
            return out[0]
        return list(zip(*out))

    def train_on_batch(self, *batch):
        xys = [xy for config in map(self._read_extra, *batch) for xy in config]
        if not xys: return
        xs_ys = map(self.batch2np, zip(*xys))
        self.model.train_on_batch(*xs_ys)

    def _read_extra(self, y, _y, _x=None, x=None, extra=None):
        if np.array_equal(y,_y) and (_x is None or np.array_equal(x, _x)):
            return
        if extra is None:
            # standard loss (with mean over batch)
            yield _x, y
            return
        # custom loss (with sum over batch, and gradient normalization)
        alpha = extra.get('alpha', 1.0) / self.gradient_normalization
        if x is None: x = _x
        # penalize the prediction
        yield _x, np.insert(_y, 0, -alpha)
        # reward the reference
        yield x, np.insert(y, 0, alpha)
