#!/usr/bin/env python3

import abc
import random
import logging

from collections import defaultdict
from copy import copy

import sys, pickle
import timeit

import itertools

from structured_prediction.AbstractStructuredPredictor import Callback

from dparser.AbstractParser import SHIFT, RIGHT, LEFT, REDUCE, INVERT_MOVE
def ddisplay(derivation, n=6): return ' '.join(['\t']*n + ['{}[{:.0f}]'.format(INVERT_MOVE[move],s) for move,s in derivation])

# logging switches
transition_log = False
trace_derivation = False
uas_changes = False

# algorithmic switch
neural_mode = False

logging.basicConfig(level=logging.DEBUG)

from misc.io import Listener

################################################################################
##                               DYNAMIC CLASSES                              ##
################################################################################

def in_scope(cls):
    cls.__module__ = __name__
    classes[cls.__name__] = cls
    globals()[cls.__name__] = cls
    return cls

def system(system, clas, path=''):
    from importlib import import_module
    if path: path += '.'
    return getattr(import_module(path+system), clas)

class classes(dict):
    def __missing__(self, name):
        implem = 'BeamParser' if 'BeamParser' in name else 'LocalParser'
        output = system(('labeled_' if 'Labeled' in name else '')+('neural' if neural_mode else 'perceptron')+'_parser', ('Labeled' if 'Labeled' in name else '')+('FeedForward' if neural_mode else 'Perceptron')+'Parser', path='dparser')
        s = name.replace('Beam','').replace('Local','').replace('Labeled','').replace('Parser','')
        return in_scope(system(s, implem, path='transition_system') + output)
classes = classes()

################################################################################
##                                MAIN INTERFACE                              ##
################################################################################

def init_dependency_parser(system='ArcEager', **params):
    features_params = ['delexicalized', 'embeddings']
    features_params = {name:v for name, v in params.items() if name in features_params}
    if neural_mode:
        from dparser.features import dict_features as feature_templates
    else:
        from dparser.features import zn11_features as feature_templates
    features = feature_templates
    if features_params:
        from functools import partial
        features = partial(features, **features_params)
    parser = classes[system+("Beam" if params.get('beam', 8) else "Local")+'Parser'](features)
    for param, value in params.items():
        if param in features_params: continue
        if hasattr(parser, 'with_'+param):
            parser = getattr(parser, 'with_'+param)(value)
        else:
            setattr(parser, param, value)
    return parser

def train_dependency_parser(parser, dataset, n_epoch,
                            listener=Listener, compute_coverage=False, feature_file=None,
                            strategy=21,
                            embeddings=None,
                            p_no=(0,0,0,0),
                            iter_callback=None, crossvalidate=False,
                            average=True, shuffle=True, code_injection=None, **kwargs):
    """
    Trains a dependency parser.

    Arguments
    ---------
    - parser: an instance of Parser, or None, or a dict
        The dependency parser to train
        If None or a dict, XXX
    - dataset: a list of Pan4 tuples with bottom-up representations
        The training set
    - n_epoch: an integer
        The number of epochs to perform (the upper limit, in case of early stopping)

    Keyword arguments
    -----------------
    - listener: an object implementing a `msg` and `iter` functions (default: XXX)
        a function called to monitor training
    - strategy: anything, typically an int or a str (default: 21)
        XXX
    - shuffle: a boolean (default: True)
        XXX
    - average: a boolean (default: True)
        XXX
    - iter_callback: a callable with a single `parser` parameter (default: None)
        A function to call at the end of each epoch.
    - crossvalidate: a boolean, or a int (default: False)
        XXX
    - code_injection: a callable XXX (default: None)
        XXX
    - batch_sentence: a boolean (default: False)
    - p_no: a 4-uple of floats (default: (0,0,0,0))
    - compute_coverage: a boolean (default: False)
    - feature_file: a filehandler, or None (default: None)
        cf doc
    """
    # ROOT token supposed at the end

    if parser is None: parser = {}
    if type(parser) == dict:
        parser = init_dependency_parser(**parser)
    beam = getattr(parser, 'beam', None)

    def reshape(sample, itn):
        if random.random() < p_no[0]: return
        heads, labels = sample[2:]
        if type(heads) == tuple:
            heads, constraints = heads[:2]
        else:
            constraints = None
        heads = [None if random.random() < p_no[1] else h for h in heads]
        if constraints: heads = heads, constraints
        return sample[:2] + (heads, labels)



    #############################
    ########## LOGGING ##########    & initialize(dataset) & dump_features
    #############################

    from threading import Lock
    logger = type('', (), {'lock': Lock()})()

    def on_train_begin(parser, dataset):
        parser.initialize(dataset, embeddings=embeddings)
        logger.n_updates = 0
        parser.curve = []
    def on_epoch_begin(parser, itn):
        listener.msg("start epoch n°{}".format(itn + 1))
        logger.correct = logger.predicted = logger.total = 0
        logger.correct_pass1 = logger.total_pass1 = 0
        logger.final_configs = 0
        logger.out_correct = logger.out_constraints = logger.out_total = 0
        logger.n_sentences = 0
        #
        if itn == 9:
            parser.model.dump_features(feature_file)
        if itn == 10:
            parser.model.dump_features(None)
    def on_epoch_end(parser, itn):
        listener.msg("")
    def on_sample_begin(parser, sample):
        gold_heads, constraints = sample[2], None
        if type(gold_heads) == tuple: gold_heads, constraints = gold_heads
        out_constraints = sum(1 for h in constraints if h) if constraints else 0
        with logger.lock:
            logger.out_constraints += out_constraints
    def on_sample_end(parser, sample, before, did_update, final_config):
        gold_heads = sample[2]
        if type(gold_heads) == tuple: gold_heads = gold_heads[0]
        if beam:
            predicted = final_config.depth
            total = final_config.length_final
            out_correct = 0 #
        else:
            heads, depth = final_config
            predicted = total = depth
            out_correct = sum(1 for gh,h in list(zip(gold_heads, heads))[1:-1] if gh==h)
        out_total = sum(1 for gh in gold_heads if gh)
        with logger.lock:
            logger.predicted += predicted
            logger.total += total
            logger.out_correct += out_correct
            logger.out_total += out_total
            logger.n_sentences += 1
    def on_timestep(suboptimal, config, XXX):
        pass
    def on_update_begin_learn(suboptimal, y, _y, sample):
        pass
    def on_update_end_learn(suboptimal, config):
        update_depth = config[-1]
        with logger.lock:
            if not suboptimal: logger.total_pass1 += update_depth
            logger.n_updates += 1
    def on_update_infer(full_search):
        if full_search:
            with logger.lock:
                logger.final_configs += 1
    def after_epoch_eval(parser, itn, perf):
        if type(perf) in [int, float]:
            listener.msg("Score at iteration {}: {:.3%}".format(itn + 1, perf))
        elif type(perf) == tuple and perf and all(type(p) in [int, float] for p in perf):
            listener.msg("Scores at iteration {}: {}".format(itn + 1, ' - '.join("{:.2%}".format(p) for p in perf)))
        elif type(perf) == tuple and perf and all(type(p[1]) in [int, float] for p in perf):
            listener.msg("Scores at iteration {}: {}".format(itn + 1, ' - '.join(("[{}] {:.2%}" + " ({})" if len(p)>2 else "").format(*p) for p in perf)))
        parser.curve.append((itn + 1, logger.n_updates, perf))



    ########################################
    ########## OPTIONAL BEHAVIORS ##########
    ########################################

    ### Display updated derivations
    if transition_log:

        def on_sample_begin(parser, sample, on_sample_begin=on_sample_begin):
            on_sample_begin(parser, sample)
            listener.msg("\n{}\n{}".format(*(sample[:2])))
        def on_update_begin_learn(suboptimal, y, _y, sample, on_update_begin_learn=on_update_begin_learn):
            on_update_begin_learn(suboptimal, y, _y, sample)
            listener.msg("\nn0\tguess\t\tbest" if beam else "n0\tguess\t\tbest\t\tstack")
            listener.msg("----------------------------")
        def on_timestep(suboptimal, config, XXX, on_timestep=on_timestep):
            on_timestep(suboptimal, config, XXX)
            y, _y, _X, X = config[:4]
            parse, scores = XXX
            action_pattern = "{} [{:.2f}]"
            if beam:
                from dparser.AbstractBeamParse import ID
                depth_description = "{}\t{}\t{}".format(parse.n0[ID], action_pattern, action_pattern if y!=_y else "." if X!=_X else "").format(
                    INVERT_MOVE[_y], scores[_y], INVERT_MOVE[y], scores.get(y,0))
            else:
                n0 = parse
                depth_description = "{}\t{}\t{}\t{}".format(n0, action_pattern, action_pattern, stack).format(
                    INVERT_MOVE[_y], scores[_y], INVERT_MOVE[y] if y!=_y else "\t", scores[y] if y!=_y else 0)
            listener.msg(depth_description)

    ### Force full search to compute exact accuracies
    if compute_coverage:

        def on_timestep(suboptimal, config, XXX, on_timestep=on_timestep):
            on_timestep(suboptimal, config, XXX)
            y, _y, _X, X = config[:4]
            if y==_y and X==_X:
                with logger.lock:
                    if not suboptimal: logger.correct_pass1 += 1
                    logger.correct += 1
        def on_epoch_end(parser, itn, on_epoch_end=on_epoch_end):
            on_epoch_end(parser, itn)
            # Gold transitions == at first pass
            if beam: listener.msg(
                "Gold transitions (train) at iteration {}: {:.3%}".format(itn + 1, logger.correct / logger.total if logger.total else 1))
            else: listener.msg(
                "UAS (train) at iteration {}: {:.3%}".format(itn + 1, logger.out_correct / logger.out_total if logger.out_total else 1))
            if logger.final_configs:
                #
                listener.msg(
                    "Transitions seen on first update:\t{:.3%}".format(logger.total_pass1 / logger.total if logger.total else 1))
                listener.msg("Regstops:\t{}\t/\t{}".format(logger.final_configs, logger.n_sentences))

    ### Compare accuracies just before and just after an update
    if uas_changes:

        def on_epoch_begin(parser, itn, on_epoch_begin=on_epoch_begin):
            on_epoch_begin(parser, itn)
            global uass
            uass = []
        def on_sample_begin(parser, sample, on_sample_begin=on_sample_begin):
            # CAUTION: ignore any previously returned value
            on_sample_begin(parser, sample)
            return next(parser.process_sample(sample))
        def on_update_begin_learn(suboptimal, y, _y, sample, on_update_begin_learn=on_update_begin_learn):
            on_update_begin_learn(suboptimal, y, _y, sample)
            words = sample[0]
            heads = sample[2]
            if type(heads) == tuple: heads = heads[0]
            if len(words) < 10:
                listener.msg("++++++++++++++++++++++++++++++++++++++++")
                from pan4.io import display_tree
                listener.msg(display_tree(words, heads))
                listener.msg("C- {}".format(_y.score))
                listener.msg(ddisplay(_y.derivation()))
                listener.msg(display_tree(words, _y.heads))
                listener.msg("C+ {}".format(y.score))
                listener.msg(ddisplay(y.derivation()))
                listener.msg(display_tree(words, y.heads))
        def on_sample_end(parser, sample, before, did_update, final_config, on_sample_end=on_sample_end):
            on_sample_end(parser, sample, before, did_update, final_config)
            heads = sample[2]
            if type(heads) == tuple:
                gold_heads, constraints = heads[:2]
            else:
                gold_heads, constraints = heads, None

            if did_update:
                heads_before, (score_before, d_before) = before
                heads_after, (score_after, d_after) = next(parser.process_sample(sample))
                uas_before = sum(1 for b,g in zip(heads_before,gold_heads) if b==g and g!=None)/len(gold_heads[1:])
                uas_after = sum(1 for b,g in zip(heads_after,gold_heads) if b==g and g!=None)/len(gold_heads[1:])
                uass.append(100 * (uas_after - uas_before))

#                """
                from pan4.io import display_tree
                words, tags = sample[:2]
                if len(words) < 10:
                    listener.msg("----------------------------------------")
                    listener.msg("")
                    listener.msg(display_tree(words, gold_heads, data2=tags))
                    listener.msg("<<<<< {:.1%}\t{}\t{}".format(uas_before, sum(1 for h,g in zip(heads_before,gold_heads) if h==g and g), score_before))
                    listener.msg(display_tree(words, heads_before, data2=tags))
                    listener.msg(str(sum(1 for h,g in zip(heads_before,gold_heads) if h==g and g)) + '\t' + ddisplay(d_before))
                    listener.msg(str(sum(1 for h,g in zip(heads_after,gold_heads) if h==g and g)) + '\t' + ddisplay(d_after))
                    listener.msg(display_tree(words, heads_after, data2=tags))
                    listener.msg(">>>>> {:.1%}\t{}\t{}".format(uas_after, sum(1 for h,g in zip(heads_after,gold_heads) if h==g and g), score_after))
#                """
        def on_epoch_end(parser, itn, on_epoch_end=on_epoch_end):
            on_epoch_end(parser, itn)
            global uass
            from misc.stats import stats, stats_middle
            stats("Delta UAS", uass)
            stats_middle("Mid-D UAS", uass)
            listener.msg("< 0: {:.2%}\n= 0: {:.2%}\n> 0: {:.2%}".format(sum(1 for u in uass if u<0)/len(uass), sum(1 for u in uass if u==0)/len(uass), sum(1 for u in uass if u>0)/len(uass)))



    cb = Callback()

    cb.on_train_begin = on_train_begin
    cb.on_epoch_begin = on_epoch_begin
    cb.on_epoch_end = on_epoch_end
    cb.on_sample_begin = on_sample_begin
    cb.on_sample_end = on_sample_end

    cb.shuffle = shuffle

    cb.epoch_eval = iter_callback
    cb.before_epoch_eval = lambda *_: listener.msg("Callback...")
    cb.after_epoch_eval = after_epoch_eval

    cb.polish = average
    cb.on_polish = lambda *_: listener.msg('Averaging weights')

    cb.on_timestep = on_timestep
    cb.on_update_begin_learn = on_update_begin_learn
    cb.on_update_end_learn = on_update_end_learn
    cb.on_update_infer = on_update_infer

    if code_injection: cb, reshape = code_injection(cb, reshape, logger)

    #########################################
    ########## TRAINING STRATEGIES ##########
    #########################################

    def _strategy(strategy, n_epoch):
        #
        st = dict(kwargs)

        st.update(dict(zip('time_limit check_cut_edge check_single_reference oracle restart exploration_rate beam_update remember_bad_beams'.split(), strategies(parser, strategy))))

        ###
        st['compute_coverage'] = compute_coverage
        st['p_no'] = p_no
        return st

    def _epoch_strategy(strategy, itn, n_epoch):
        st = {}
        exploration_rate = exploration_strategy(strategy, itn, n_epoch)
        if exploration_rate is not None:
            st['exploration_rate'] = exploration_rate
        ###
        st['feature_file'] = feature_file if itn==9 else None
        return st

    def _sample_strategy(strategy, sample, itn, n_epoch):
        st = {}
        if strategy==10:
            words = sample[0]
            length_final = parser._initBeamParse(words).length_final if beam else len(words[1:-1])
            import math
            time_limit = max(5,math.ceil(length_final*(itn+1)/n_epoch))
            st['time_limit'] = time_limit
        return st


    #
    parser.strategy = _strategy
    parser.epoch_strategy = _epoch_strategy
    parser.sample_strategy = _sample_strategy

    if crossvalidate: crossvalidate = 2.5
    parser.train(dataset, n_epoch, strategy=strategy, listener=listener, feature_file=feature_file, crossvalidate=crossvalidate, reshape=reshape, cb=cb, batched=(lambda itn: 2**itn) if neural_mode else None)

    #
    del parser.strategy, parser.epoch_strategy, parser.sample_strategy

    return parser

def process_dependency_parser(parser, dataset, listener=Listener, test=False, **kwargs):
    dataset = listener.iter(dataset)
    if not parser: return dataset

    def postprocess(ex, heads):
        labels = [None for _ in ex[0]]
        if type(heads) == tuple: heads, labels = heads
        if test:
            gold_heads, gold_labels = ex[2:]
            heads = (gold_heads[:2] if type(gold_heads) == tuple else (gold_heads,)) + (heads,)
            labels = (gold_labels[:2] if type(gold_labels) == tuple else (gold_labels,)) + (labels,)
        return ex[:2] + (heads, labels)

    if trace_derivation:
        kwargs['return_history'] = "scores"
    return parser.process(dataset, postprocess=postprocess, batched=1000 if neural_mode else None, **kwargs) ###


################################################################################
##                            FINE GRAIN EVALUATION                           ##
################################################################################

#
def eval_read(testset, listener=Listener):

    def read_test(gold, pred):
        from conll.io import rt
        #
        with open(gold, "rt", encoding="utf-8", errors="replace") as gold, open(pred, "rt", encoding="utf-8", errors="replace") as pred:
            gold = rt(gold, multiroots=True)
            pred = rt(pred, multiroots=True)
            for g,p in zip(gold,pred):
                while g[0] != p[0]: g = next(gold)
                (w,t,gh,gl),(_,_,h,l) = g,p
                yield w,t,(gh,h),(gl,l)

    def read_trans(gold, trans):
        from conll.io import rt
        with open(gold, "rt", encoding="utf-8", errors="replace") as gold, open(trans, "rt", encoding="utf-8", errors="replace") as trans:
            gold = rt(gold, multiroots=True)
            trans = rt(trans, multiroots=True)
            for g,p in zip(gold,trans):
                (gw,t,gh,l),(w,_,_,_) = g,p
                #
                win = 2
                h = [gh if gw in w[max(i-win,0):i+win+1] else [gh] for i,(gw,gh) in enumerate(zip(gw,gh))]
                yield gw,t,(gh,h),l

    def read_binvalues(testset):
        from conll.io import rt
        with open(testset, "rt", encoding="utf-8", errors="replace") as testset:
            testset = rt(testset, multiroots=True)
            for w,t,(gh,b),_ in testset:
                h = [gh if b else [gh] for b,gh in zip(b,gh)]
                yield w,t,(gh,h),[None for _ in gh]

    if type(testset) == str: # binannotatedpath
        return read_binvalues(testset)
    if type(testset) == tuple:
        if testset[0] == None: # (None, goldpath, transpath)
            return read_trans(*testset[1:])
        elif type(testset[0]) == str: # (goldpath, predpath)
            return read_test(*testset)
        else: # (parser, dataset)
            return process_dependency_parser(*testset, test=True)
    if type(testset) == list: # [testset1, testset2]
        return [eval_read(t) for t in testset]
    # otherwise, assume that testset is an iterable (=> generator)
    return testset

DISPLAY, ERRDISPLAY, UAS, CONFUSION, STATS, COMPARE, COMPFUSION = range(7)

def eval_dependency_parser(testset, ignore_tags=[".","PUNCT"], ignore_labels=[], by_type=None, subset=None, listener=Listener, out=UAS):

    ignore_tags += ["<start>","ROOT"]
    testset = eval_read(testset, listener=listener)

    if out in [DISPLAY, ERRDISPLAY]:
        from pan4.io import display_tree
        from structured_prediction.AbstractStructuredPredictor import display_routine
        keep = lambda ex: len(ex[0]) < 15
        display = lambda x: display_routine(testset, keep, x)
        if out == DISPLAY:
            display(lambda ex:
                display_tree(ex[1], ex[2], ex[0]))
        if out == ERRDISPLAY:
            display(lambda ex:
                (display_tree(ex[1], [gh if gh!=h else None for gh,h in zip(ex[2][0], ex[2][-1])], ex[0]),) +
                ((listener.msgs(['','=>']),
                  display_tree(ex[1], [h if gh!=h else None for gh,h in zip(ex[2][0], ex[2][-1])], ex[0]))
                 if ex[2][0] != ex[2][-1] else tuple()) +
                (listener.msgs(['','*'*150]),))

    from structured_prediction.AbstractStructuredPredictor import contains_type
    contains_depth = contains_type(by_type, DEPTH) or contains_type(by_type, LDEPTH) or contains_type(by_type, MDEPTH) or type(subset)==tuple and (contains_type(subset[0], DEPTH) or contains_type(subset[0], LDEPTH))
    contains_freq = contains_type(by_type, FREQUENCE)

    from functools import partial
    localzip = partial(eval_localzip, contains_depth=contains_depth)
    keep = partial(eval_keep, subset=subset, ignore_tags=ignore_tags, ignore_labels=ignore_labels)
    category = eval_category

    if contains_freq:
        from itertools import tee
        from collections import Counter
        if type(testset) == list:
            tokens = []
            newtestset = []
            for testset in testset:
                testset, data = tee(testset)
                tokens += [w for ex in data for w in ex[0] if w not in ["<start>","ROOT"]]
                newtestset.append(testset)
            testset = newtestset
        else:
            testset, data = tee(testset)
            tokens = [w for ex in data for w in ex[0] if w not in ["<start>","ROOT"]]
        n_tokens = len(tokens)
        zipf = Counter(tokens).most_common()
        n_types = len(zipf)
        frequences = {w: (occ/n_tokens, rank, n_types) for rank,(w,occ) in enumerate(zipf, start=1)}
        category = partial(eval_category, frequences=frequences)

    def config(tok): # _x, _y, y
        _, w, _, h, _, _ = tok
        return w, h[-1], h[0]

    from structured_prediction.AbstractStructuredPredictor import eval_routine, matrix_routine, stats_routine, compare_routine, matrix_compare_routine
    routine = eval_routine if out == UAS else matrix_routine if out == CONFUSION else stats_routine if out == STATS else compare_routine if out == COMPARE else matrix_compare_routine if out == COMPFUSION else None
    if routine:
        return routine(testset, localzip, config, keep, category, eval_type=by_type)


#
def eval_localzip(ex, contains_depth=False):
    words, tags, heads, labels = ex
    tags = [t.cpos if 'cpos' in dir(t) else t for t in tags]
    from pan4.preprocess import depths
    gh = heads[0] if type(heads) == tuple else heads
    if not heads or type(heads) == tuple and not heads[-1]: return tuple()
    depths = depths(gh) if contains_depth else [None for _ in gh]
    if type(heads) == tuple: heads = zip(*heads)
    if type(labels) == tuple: labels = zip(*labels)
    return ((i,)+tok for i,tok in enumerate(zip(words, tags, heads, labels, depths)))

def eval_keep(tok, ex=None, subset=None, ignore_tags=[], ignore_labels=[]):
    if not tok: return False
    i,w,t,h,l,d = tok
    if t in ignore_tags: return False
    if (l[0] if type(l) == tuple else l) in ignore_labels: return False
#    if type(h) == tuple and len(h) == 3 and h[1] is not None: return False
    if type(h) == tuple and len(h) == 3 and h[1] is not None and h[1]==h[2]: return False
    #
    if type(h) == tuple and h[-1] is None or h is None: return False
    if subset is not None: return eval_category(tok, ex, subset[0]) == subset[1]
    return True

WORD, POS, LABEL, DEPTH, MDEPTH, LDEPTH, LENGTH, MLENGTH, LLENGTH, HEADPOS, DIRECTION, QUARTER, ISROOT, FREQUENCE = range(1,15)

def eval_category(tok, ex, by_type, frequences=None, gold=True):
    if type(by_type) == tuple:
        return tuple(eval_category(tok, ex, sub_type, frequences=frequences, gold=gold) for sub_type in by_type)
    if type(by_type) == list:
        return "".join(str(eval_category(tok, ex, sub_type, frequences=frequences, gold=gold)) for sub_type in by_type)
    import math
    tags = ex[1]
    i,w,t,h,l,d = tok
    if type(h) == tuple: h = h[0 if gold else -1]
    if type(l) == tuple: l = l[0]
    if by_type == WORD:
        return w
    if by_type == POS:
        return t
    if by_type == LABEL:
        return l
    if by_type == DEPTH:
        return d
    if by_type == MDEPTH:
        return min(d, 4) if d is not None else None
    if by_type == LDEPTH:
        return 2**int(math.log(d,2)) if d is not None else None
    if by_type == LENGTH:
        return abs(h-i) if None not in [h,i] else None
    if by_type == MLENGTH:
        return min(abs(h-i), 4) if None not in [h,i] else None
    if by_type == LLENGTH:
        return 2**int(math.log(abs(h-i),2)) if None not in [h,i] else None
    if by_type == HEADPOS:
        if h is None: return None
        return tags[h]
    if by_type == DIRECTION:
        if h is None: return ""
#         if tags[h]=="ROOT": return ""
        if h>i: return "+"
        if h<i: return "-"
        return ""
    if by_type == QUARTER:
        return "Q" + str(int(4*(i-1)/len(tags[1:-1])))
    if by_type == ISROOT:
        if tags[h]=="ROOT": return "R"
        return "NR"
    if by_type == FREQUENCE:
        freq, rank, n_types = frequences[w]
        return int(math.log(n_types/rank,2))
    return by_type

################################################################################
##                                I/O UTILITIES                               ##
################################################################################

def read_conll(filehandler, max_sent=None, multiroots=False, projective_only=False):
    from conll.io import rt
    #
    nofilter = rt(filehandler, max_sent, multiroots)
    if projective_only:
        return filter(lambda s: check_projectivity(s[2]), nofilter)
    return nofilter

def read_tiger(filehandler, max_sent=None):
    from tiger.io import rt
    return rt(filehandler, max_sent)

################################################################################
##                                 TREE MANIPS                                ##
################################################################################

from pan4.preprocess import check_projectivity

from dparser.mst import maximum_spanning_tree

################################################################################
##                             BUILT-IN STRATEGIES                            ##
################################################################################

def strategies(parser, strategy):
    from structured_prediction.AbstractBeamStructuredPredictor import EARLY_DET, EARLY_LONGEST_PREFIX, EARLY_TOP_SIBLING, EARLY_TOP_SIBLING_PAST, MAX_VIOLATION_DET, MAX_VIOLATION_FROM_INIT, MAX_VIOLATION_FROM_TOP_SIBLING, MAX_VIOLATION_ND, DYNAMIC_ALONG

    time_limit = None
    check_cut_edge = True
    check_single_reference = False
    oracle = EARLY_TOP_SIBLING
    restart = False
    exploration_rate = 1
    beam_update = False
    remember_bad_beams = 0

    if strategy=="suboptimal":
        import structured_prediction.AbstractBeamStructuredPredictor as ABSP
        ABSP.p_reset = 0.003
        strategy = "earlyND"

    # main strategies
    if strategy==20:
        strategy = "greedyn"
    elif strategy==21:
        strategy = "earlyND"
    elif type(strategy) != str:
        strategy = str(strategy)


    if 'greedy' in strategy or 'goldberg' in strategy:
        parser.beam = 1
        oracle = EARLY_TOP_SIBLING
        restart = True
        if strategy=="greedyND":
            pass #exploration_rate = 0
        elif strategy=="greedy":
            oracle = EARLY_DET
            check_cut_edge = False
            check_single_reference = True
            #exploration_rate = 0
        elif strategy=="greedyn":
            pass
        elif strategy=="goldberg":
            pass #exploration_rate = 0 if itn < 1 else 0.9

    if strategy=="full":
        oracle = EARLY_DET
        check_cut_edge = False

    if 'early' in strategy or 'maxv' in strategy:

        if strategy in ['early', 'maxv']:
            check_cut_edge = False
            check_single_reference = True
        restart = 'restart' in strategy
        if restart and 'SO' not in strategy:
            pass #exploration_rate = 0

        if strategy=="early":
            oracle = EARLY_DET
        elif "earlyND" in strategy: # earlyND, earlyNDrestart, earlyNDrestartSO
            oracle = EARLY_TOP_SIBLING
        if strategy=="maxv":
            oracle = MAX_VIOLATION_DET
        elif "maxvND" in strategy: # maxvND, maxvNDrestart, maxvNDrestartSO
            oracle = MAX_VIOLATION_ND

    ################
    ### VARIANTS ###
    ################

    # time horizon
    if strategy==9:
        check_cut_edge = False
        time_limit = 10
        oracle = EARLY_LONGEST_PREFIX
        restart = True
    elif strategy==10:
        check_cut_edge = False
        #time_limit = max(5,math.ceil(length_final*(itn+1)/n_epoch))
        oracle = EARLY_LONGEST_PREFIX

    # searn
    if strategy==26:
        oracle = EARLY_TOP_SIBLING
        restart = True
        #exploration_rate = itn/n_epoch
    elif strategy==27:
        oracle = MAX_VIOLATION_FROM_INIT
        restart = True
        #exploration_rate = itn/n_epoch
    elif strategy==29:
        oracle = MAX_VIOLATION_FROM_INIT
        restart = True
        #exploration_rate = 2*itn/n_epoch

    # error detection
    elif strategy==31: # beam ref outside first half
        oracle = EARLY_TOP_SIBLING_PAST
        remember_bad_beams = 0.5

    # whole beam update
    elif strategy in [41,44]:
        beam_update = True
        if strategy==41:
            oracle = EARLY_TOP_SIBLING
        elif strategy==44:
            oracle = MAX_VIOLATION_FROM_INIT

    if strategy=="naivedynamic":
        check_cut_edge = False
        oracle = DYNAMIC_ALONG
    elif strategy=="earlynaivedynamic":
        oracle = DYNAMIC_ALONG

    return time_limit, check_cut_edge, check_single_reference, oracle, restart, exploration_rate, beam_update, remember_bad_beams

def exploration_strategy(strategy, itn, n_epoch):
    if strategy in ["goldberg"]:
        return 0 if itn < 1 else 0.9
    if strategy in [23,"earlyNDrestart","maxvNDrestart","greedyND","greedy"]:
        return 0
    if strategy in [26,27]:
        return itn/n_epoch
    if strategy==29:
        return 2*itn/n_epoch

if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--system', type=str, default='ArcEager')
    parser.add_argument('--strategy', type=str, default='greedyn')
    parser.add_argument('--epoch', type=int, default=20)
    parser.add_argument('--trainset', type=argparse.FileType('rt'))
    parser.add_argument('--trainsize', type=int, default=None)
    parser.add_argument('--devset', type=argparse.FileType('rt'))
    parser.add_argument('--testset', type=argparse.FileType('rt'))
    parser.add_argument('--model', type=str)
    parser.add_argument('--train', action='store_true')
    parser.add_argument('--validate', action='store_true')
    parser.add_argument('--eval', action='store_true')
    parser.add_argument('--input', type=argparse.FileType('rt'))
    parser.add_argument('--output', type=argparse.FileType('wt'))
    parser.add_argument('--showprogress', action='store_true')

    args = parser.parse_args()

    try: strategy = int(args.strategy)
    except: strategy = args.strategy
    n_epoch = 5 if args.devset and args.validate else args.epoch

    if args.showprogress:
        from misc.io import FrogressListener
        listener = FrogressListener
    else:
        listener = Listener

    import pickle
    if args.train:
        trainset = read_conll(args.trainset, multiroots=True)
        if args.trainsize is not None:
            trainset = list(trainset)
            random.shuffle(trainset)
            trainset = trainset[:args.trainsize]
        if args.devset:
            devset = list(read_conll(args.devset, multiroots=True))
            def iter_callback(parser):
                return eval_dependency_parser((parser, devset))
        else:
            iter_callback = None
        model = train_dependency_parser(dict(system=args.system), trainset, n_epoch, listener=listener, strategy=strategy, iter_callback=iter_callback, crossvalidate=iter_callback and args.validate)
        if args.model:
            with open(args.model, "wb") as fh:
                pickle.dump(model, fh, pickle.HIGHEST_PROTOCOL)
    else:
        try:
            with open(args.model, "rb") as fh:
                model = pickle.load(fh)
        except AttributeError as e:
            import re
            classes[re.match("Can't get attribute '([^']*Parser)'", e.args[0]).group(1)]
            with open(args.model, "rb") as fh:
                model = pickle.load(fh)

    if args.eval:
        testset = list(process_dependency_parser(model, read_conll(args.testset, multiroots=True), listener=listener, test=True))
        if args.showprogress: print()
        #print('UAS: {:.2%}'.format(eval_dependency_parser(iter(testset))))
        ignore_tags = ['.', 'PUNCT'] + ['<start>','ROOT']
        ignore_labels = []
        pairs = [((gh,gl),(h,l)) for ex in testset for t,gh,h,gl,l in zip(ex[1],*(ex[-2]+ex[-1])) if gh and gl and t.cpos not in ignore_tags and l not in ignore_labels]
        uas = sum(1 for ghl,hl in pairs if ghl[0]==hl[0])/len(pairs)
        lacc = sum(1 for ghl,hl in pairs if ghl[1]==hl[1])/len(pairs)
        las = sum(1 for ghl,hl in pairs if ghl==hl)/len(pairs)
        if lacc:
            print("UAS: {:.2%} -- LAS: {:.2%} -- LAcc: {:.2%}".format(uas, las, lacc))
        else:
            print("UAS: {:.2%}".format(uas))

    if args.input and args.output:
        dataset = read_conll(args.input, multiroots=True)
        from conll.io import wt, feat2col
        wt(args.output, feat2col(process_dependency_parser(model, dataset)))
