import abc

class AbstractBeamParse(metaclass=abc.ABCMeta):

    def __init__(self,
                 _prev=(None,None,0)
                 ):
        self.previous, self.last_move, move_score = _prev
        self.score = (self.previous.score if self.previous else 0) + move_score
        if self.previous:
            self.sentence = self.previous.sentence
            self.constraints = self.previous.constraints
            self.gold = self.previous.gold
        self.depth = self.previous.depth + 1 if self.previous else 0
        # useful API
        self.mark = False
        self.alive = True
        # extracted features, gold moves, scores
        self.configuration = None, None, None

    @abc.abstractmethod
    def __str__(self):
        pass

    @abc.abstractmethod
    def __eq__(self, other):
        pass

    def __ne__(self, other):
        return not self.__eq__(other)

    @abc.abstractmethod
    def same(self, other):
        pass

    @property
    @abc.abstractmethod
    def final(self):
        pass

    @property
    @abc.abstractmethod
    def length_final(self):
        pass

    @property
    def predecessors(self):
        sub_parse = self
        while sub_parse:
            yield sub_parse
            sub_parse = sub_parse.previous

    @abc.abstractmethod
    def valid_moves(self):
        pass

    def cost(self, move, constraints=False):
        return sum(self.get_cut_edges(move, constraints))

    def no_cost(self, move, constraints=False):
        return next(self.get_cut_edges(move, constraints), None) is None

    def min_cost(self, moves, constraints=False):
        costs = [(move, self.get_cut_edges(move, constraints)) for move in moves]
        moves = []
        while not moves:
            moves = [move for move, cost in costs if next(cost, None) is None]
        return moves

    @abc.abstractmethod
    def get_cut_edges(self, move, constraints=False):
        pass

    @abc.abstractmethod
    def transition(self, move, score=0):
        pass

    def transitions(self, *moves):
        return sum(self, *moves)

    def __add__(self, move):
        if type(move) == list:
            return self.transitions(*move)
        if type(move) == tuple:
            return self.transition(*move)
        return self.transition(move)

    def __radd__(self, move):
        return self

    @property
    @abc.abstractmethod
    def heads(self):
        pass

    ###
    def set_label(self, labels):
        last_label = self.__dict__.get('last_label')
        if type(last_label) == tuple:
            labels[last_label[0]] = last_label[1]
            return True
        return False

    @property
    def derivation(self):
        derivation = [(parse.last_move, parse.score - parse.previous.score) for parse in self.predecessors if parse.last_move is not None]
        derivation.reverse()
        return derivation
