import abc
import random, itertools, pickle

from .AbstractParser import AbstractParser
from structured_prediction.AbstractBeamStructuredPredictor import AbstractBeamStructuredPredictor
from structured_prediction.AbstractStructuredPredictor import Callback
from misc.itertools import next0, catch0
from pan4.format import root_at_start, root_at_end

from structured_prediction.AbstractBeamStructuredPredictor import DYNAMIC_ALONG

# algorithmic switch
mixture_update = False

def mixture(xs):
    xs = list(xs)
    s = sum(sorted(xs))
    return [x/s for x in xs]

def exp_mixture(xs):
    xs = list(xs)
    m = max(xs)
    import math
    return mixture(math.exp(x - m) for x in xs)

class AbstractBeamParser(AbstractBeamStructuredPredictor, AbstractParser):

    @abc.abstractmethod
    def _defaultBeamParse(self):
        pass

    def _initBeamConfig(self, sample, config=None):
        if not config: config = self._defaultBeamParse()
        config.sentence, config.gold, config.constraints = self._init_config(sample, self.root_first)
        return config

    def search(self, blackbox, sample, return_all=True, return_history=None, forced_decoding=False, cb=Callback()):
        """Beam search decoding for given sentence, with current model

        Tree computation is lazy, which means that when used for 1-best decoding, only one derivation is converted into a list of heads.

        Beam search decoding follows [Zhang and Clark, 2008].

        [Zhang and Clark, 2008] A Tale of Two Parsers: Investigating and Combining Graph-based and Transition-based Dependency Parsing Using Beam-search, EMNLP'08

        Arguments
        ---------
        - sample: a Pan4 tuple with bottom-up representations
            The tokenized input sentence.
            The first or unique `heads` information is ignored, and internally forced to None.
            If (partial) constraints are specified, the not-None heads must be included in the final output (search space is restrained accordingly).

        Keyword arguments
        -----------------
        - return_all: a boolean (default: True)
            If False, reduce the k-best hypotheses by eliminating duplicate parse trees issued by distinct derivations (keeping the best scored one).

        Global controls
        ---------------
        - dependency_parser.trace_derivation: a boolean
            Enables output of the transition sequence for each hypothesis.
            This requires an extra computation time, linear in the derivation length.

        Returns
        -------
        - a generator yielding hypotheses as `(heads, (score, derivation))`:
            - heads: a list of None and ints
                The predicted parse tree of the input sentence for this derivation hypothesis.
            - score: a float
                The global score of the derivation (the sum of transition scores).
            - derivation: a list of (int, float) tuples, or None
                If `trace_derivation` enabled, the list of transitions in the derivation, along with each transition's score (without the initial SHIFT if parsed with the 'ROOT' token at the end); None otherwise.
            Hypotheses are yielded in decreasing order of score.
            The generator yields at least one value and at most `self.k`.
        """

        if not forced_decoding:
            sample = self.forget_gold(sample)

        kbest, _ = yield from next0(self._infer(blackbox, sample, forced_decoding=forced_decoding))
        heads = []

        for config in kbest:
            parse_heads = config.heads

            if self.root_first:
                root_at_end(heads=parse_heads)

            # choice of derivation with highest score
            if return_all or parse_heads not in heads:
                heads.append(parse_heads)
                history = None
                if return_history == "scores": history = config.score, config.derivation()
                if return_history == "display": history = config.format_derivation()
                if return_history == "configuration": history = config
                yield False, (parse_heads, history)

    def learn(self, blackbox, sample,
              oracle=None, restart=True, batch_sentence=False,
              beam_update=False, p_no=(0,0,0,0),
              compute_coverage=False, feature_file=None, cb=Callback(),
              **kwargs):
        """Online training of the model on one sentence

        Early update and max-violation training strategies follow [Huang et al., 2012].
        Dynamic oracle computation follows [Goldberg and Nivre, 2012].

        [Huang et al., 2012] Structured Perceptron with Inexact Search, NAACL'12
        [Goldberg and Nivre, 2012] A Dynamic Oracle for Arc-Eager Dependency Parsing, COLING'12

        Arguments
        ---------
        - sample: a Pan4 tuple with bottom-up representations
            The tokenized sentence used as training example.
            The first or unique `heads` information is used as the gold parse tree. It can be partial, but in this case the given forest must be compatible with a full projective tree.
            If (partial) constraints are specified, the search space is restrained to train only on configurations compatible with those heads.

        Keyword arguments
        -----------------
        - oracle: an int, or None
        - restart: a boolean (default: True)
        - batch_sentence: a boolean (default: False)
        - beam_update: a boolean (default: False)
        - p_no: a 4-uple of floats (default: (0,0,0,0))
        - compute_coverage: a boolean (default: False)
        - feature_file: a filehandler, or None (default: None)
        - exploration_rate: a float (default: 1)
        - check_cut_edge: a boolean (default: True)
        - check_single_reference: a boolean (default: False)
        - time_limit: an int, or None (default: None)
        - remember_bad_beams: a float (default: 1)
            cf doc

        Returns
        -------
        - ***

        Side-effects
        ------------
        - the weight vector of the model is updated according to its behavior on the sentence and to the specified learning strategy
        - features are dumped for each evaluated configuration, if a `feature_file` has been registered
        """
        before = cb.on_sample_begin(self, sample)

        update_states = yield from catch0(self._infer(blackbox, sample, oracle=oracle, beam_update=beam_update, feature_file=feature_file, cb=cb, **kwargs))
        if not restart:
            update_states = itertools.islice(update_states, 1)
        if batch_sentence:
            update_states = list(update_states)

        did_update = False

        t0 = 1
        for suboptimal, (kbest, reference) in enumerate(update_states):
            if random.random() < p_no[2]: continue

            #
            if mixture_update:
                if len(kbest)==1:
                    kbest += [config.previous.transition(move, config.previous.configuration[-1][move]) for config in kbest for move in config.previous.configuration[-1] if move != config.last_move]
                # e^sum/sum(e^sum) ignores the contribution of last common ancestor's score
                kbest_mixture = list(zip(exp_mixture(p.score for p in kbest), kbest))
            else:
                rank = -1 if beam_update else 0
                if kbest[rank].final: rank = 0
                kbest_mixture = [(int(i==rank),p) for i,p in enumerate(kbest)]

            for mixture_weight, prediction in kbest_mixture:
                # no useless traversal
                if not mixture_weight:
                    continue

                yield from self._joint_learn(prediction)

                # no false update
                if prediction.score - reference.score < 0:
                    continue

                cb.on_update_begin_learn(suboptimal, reference, prediction, sample)

                global_features = []

                for config, best_config in zip(prediction.predecessors, reference.predecessors):
                    # reached a start configuration
                    if config.last_move is None or config.mark:
                        break
                    # no more update to come... (at least if best_config is used)
                    if config is best_config and oracle != DYNAMIC_ALONG:
                        # in that case, only bother for the initial prefix (and if you want stats)
                        if suboptimal or not compute_coverage:
                            break
                    _X, gold_moves, scores = config.previous.configuration
                    X, _, scores = best_config.previous.configuration
                    guess = config.last_move
                    # only one reference path
                    if oracle != DYNAMIC_ALONG:
                        gold_moves = [best_config.last_move]
                    best = guess if guess in gold_moves else gold_moves[0] if len(gold_moves)==1 else max(gold_moves, key=lambda move: _scores[move])
                    global_features.append((best,guess,_X,X,scores,config.depth))

                    #
                    cb.on_timestep(suboptimal, global_features[-1], (config, scores)) # scores,=global_features[4:-1] ignored

                    # now they are start configurations, let's forget the past
                    config.mark = best_config.mark = True

                if feature_file:
                    to_remove = self.k * (config.length_final-global_features[0][-1])
                    pickle.dump(to_remove, feature_file)
                for y,_y,_X,X,scores,depth in reversed(global_features):
                    #
                    for _ in range(depth-t0):
                        yield False, self._joint_wrap(4, 0, 0, None, None, {}) # dummy for time averaging
                    extra = {'time_step': False}
                    if mixture_update:
                        extra['alpha'] = mixture_weight
                    yield False, self._joint_wrap(2, y, _y, _X, X, extra)
                    t0 = depth
                    if y!=_y: did_update = True

                cb.on_update_end_learn(suboptimal, global_features[0])

        prediction = max(kbest_mixture, key=lambda x:x[0])[1]
        cb.on_sample_end(self, sample, before, did_update, prediction)
