import abc

from .AbstractParser import AbstractParser
from structured_prediction.AbstractStructuredPredictor import Callback
from misc.itertools import next0, catch0
from misc.misc import DefaultList

class Parse:

    def __init__(self, n):
        self.n = n
        self.heads = [None] * (n - 1) + [0]
        self.lefts = [DefaultList(default=0) for _ in range(n + 1)]
        self.rights = [DefaultList(default=0) for _ in range(n + 1)]

    def add(self, head, child):
        self.heads[child] = head
        if child < head:
            self.lefts[head].append(child)
        else:
            self.rights[head].append(child)

    def __str__(self):
        return "\n".join(["n={}".format(self.n),
                          "heads={}".format(self.heads),
                          "lefts={}".format(self.lefts),
                          "rights={}".format(self.rights)])


class AbstractLocalParser(AbstractParser):

    def search(self, blackbox, sample, return_all=True, return_history=None, cb=Callback()):
        sample = self.forget_gold(sample)

        score, derivation = 0, []
        for _, move, _, (_, _, parse), scores in (yield from catch0(self._infer(blackbox, sample, cb=cb))):
            if return_history == "scores":
                derivation.append((move, scores[move]))
                score += scores[move]

        history = None
        if return_history == "scores": history = score, derivation
        if return_history == "display": pass #
        if return_history == "configuration": pass #

        yield False, (parse.heads, history)

    def learn(self, blackbox, sample, batch_sentence=False, cb=Callback(), **kwargs):
        before = cb.on_sample_begin(self, sample)

        update_states = self._infer(blackbox, sample, cb=cb)
        if batch_sentence:
            update_states = list(update_states)
        suboptimal = 0
        for depth, (y, _y, _x, (n0, _, parse), scores) in enumerate((yield from catch0(update_states))):
            config = y, _y, _x, _x, depth
            ###
            cb.on_update_begin_learn(suboptimal, None, parse, sample)
            cb.on_timestep(suboptimal, config, (n0, scores))
            self.model.update(y, _y, _x)
            cb.on_update_end_learn(suboptimal, config)
            if y != _y: suboptimal += 1

        cb.on_sample_end(self, sample, before, bool(suboptimal), (parse.heads, depth))

    def _initParse(self, n):
        # Description of the state of the parser
        # --------------------------------------
        #
        # Index describing where we are in the input sentence. This
        # index allows us to access to the “buffer” used in the
        # dependency litterature (buffer = sentence[n0:])
        n0 = 2
        # Words that occur before n0 but for which we have not
        # predicted a head yet (--> partially processed words)
        stack = [1]
        # Structure to hold the partially predicted tree
        parse = Parse(n + 2)

        return n0, stack, parse

    def _infer(self, blackbox, sample, cb=Callback()):
        (words, tags, root_first, n), gold, constraints = self._init_config(sample)

        n0, stack, parse = self._initParse(n)
        # the decisions we made
        hist = []

        while not self.final(n0, n, stack):
            features = self.features((words, tags, n0, stack, parse, hist))
            valid_moves = list(self.valid_moves(n0, n, stack, parse.heads))
            yield True, (features, valid_moves)
            _X, (_, scores) = next(blackbox.next)
            guess = max(valid_moves, key=lambda move: scores[move])

            hist.append(guess)

            best = None
            if gold is not None:
                gold_moves = self.get_gold_moves(n0, n, stack, parse.heads, gold, valid_moves)
                assert gold_moves
                best = guess if guess in gold_moves else max(gold_moves, key=lambda move: scores[move])

            n0 = self.transition(guess, n0, stack, parse)

            yield False, (best, guess, _X, (n0, stack, parse), scores)

        #
        cb.on_update_infer(full_search=True)

    @abc.abstractmethod
    def transition(self, move, n0, stack, parse):
        """
        Make the action `move` by modifying the buffer (i.e. the
        position in the sentence `i`), the `stack` and updating the
        (partial) dependency tree accordingly.
        """
        pass

    @abc.abstractmethod
    def valid_moves(self, n0, n, stack, heads):
        """
        Returns the authorized moves
        """
        pass

    def get_gold_moves(self, n0, n, stack, heads, gold, valid_moves=None):
        if valid_moves is None:
            valid_moves = self.valid_moves(n0, n, stack, heads)
        return [move for move in valid_moves if next(self.get_cut_edges(move, n0, stack, heads, gold), None) is None]

    def get_cut_edges(self, move, n0, stack, heads, gold):
        s0 = stack[-1] if stack else None
        _stack = lambda: reversed(stack)
        _nohead_stack = lambda: (s for s in _stack() if heads[s] is None)
        root_first = self.root_first
        n = len(heads)
        _end_buffer = lambda n0=n0: n0 + 1 == (n if root_first else n - 1)
        return self._get_cut_edges(move, n0, s0, _stack, _nohead_stack, gold, root_first, _end_buffer)

    @abc.abstractmethod
    def _get_cut_edges(self, move, n0, s0, _stack, _nohead_stack, gold, root_first, _end_buffer):
        pass

def _get_stack_context(stack, seq):
    return tuple(seq[stack[-i]] if i - 1 < len(stack) else None for i in [1, 2, 3])

def _get_parse_context(word, deps, data):
    import itertools
    deps = deps[word] if word > -1 else []
    valency = len(deps)
    return itertools.chain([valency], (data[d] for d in deps[-2:][::-1]), [None] * (2-valency))

def extract_features(words, tags, n0, stack, parse, history):
    from .AbstractParser import pad_list

    # The name of all features follows the following pattern:
    # - a letter: s --> stack n --> buffer
    # - an integer: the position on the stack or on the buffer
    # - an optionnal letter: l = left most child, r = rightmost child
    #   or h for the head
    # - a letter: p for POS tags and w for word identity

    depth = len(stack)
    s0 = stack[-1] if depth else -1

    s0w, s1w, s2w = _get_stack_context(stack, words)
    s0p, s1p, s2p = _get_stack_context(stack, tags)

    n0w, n1w, n2w = pad_list(words[n0:])
    n0p, n1p, n2p = pad_list(tags[n0:])

    n0lv, n0lw, n0l2w = _get_parse_context(n0, parse.lefts, words)
    n0lv, n0lp, n0l2p = _get_parse_context(n0, parse.lefts, tags)

    n0rv, n0rw, n0r2w = _get_parse_context(n0, parse.rights, words)
    n0rv, n0rp, n0r2p = _get_parse_context(n0, parse.rights, tags)

    s0lv, s0lw, s0l2w = _get_parse_context(s0, parse.lefts, words)
    s0lv, s0lp, s0l2p = _get_parse_context(s0, parse.lefts, tags)

    s0rv, s0rw, s0r2w = _get_parse_context(s0, parse.rights, words)
    s0rv, s0rp, s0r2p = _get_parse_context(s0, parse.rights, tags)

    head = parse.heads[s0] if depth else None
    s0hw = words[head] if head is not None else None
    s0hp = tags[head] if head is not None else None

    head2 = parse.heads[head] if head is not None else None
    s0h2w = words[head2] if head2 is not None else None
    s0h2p = tags[head2] if head2 is not None else None

    # distance between s0 and n0
    # the distance is clipped to 5
    dist = min((n0 - s0, 5)) if depth else 0

    hist_feat = [history[-i] if i <= len(history)
                 else None for i in range(1, 9)]

    #
    n0lL = n0l2L = n0rL = n0r2L = s0lL = s0l2L = s0rL = s0r2L = None

    return (s0w, s1w, s2w,
            s0p, s1p, s2p,
            n0w, n1w, n2w,
            n0p, n1p, n2p,
            n0lv, n0lw, n0l2w, n0lp, n0l2p, n0lL, n0l2L,
            n0rv, n0rw, n0r2w, n0rp, n0r2p, n0rL, n0r2L,
            s0lv, s0lw, s0l2w, s0lp, s0l2p, s0lL, s0l2L,
            s0rv, s0rw, s0r2w, s0rp, s0r2p, s0rL, s0r2L,
            s0hw, s0hp,
            s0h2w, s0h2p)

def morpho_features(words, ctx_features, n0, stack, parse, history):

    def add(name, *args):
        features[name.format(*args)] = 1

    features = defaultdict(int)

    morpho = {
        "s0": ctx_features["morpho"][stack[-1]] if len(stack) >= 1 else defaultdict(lambda: "__EMPTY__"),
        "s1": ctx_features["morpho"][stack[-2]] if len(stack) >= 2 else defaultdict(lambda: "__EMPTY__"),
        "b0": ctx_features["morpho"][n0],
        "b1": ctx_features["morpho"][n0 + 1] if len(words) > n0 + 1 else defaultdict(lambda: "__EMPTY__")
        }

    pos = {
        "s0": ctx_features["pos"][stack[-1]] if len(stack) >= 1 else "__EMPTY__",
        "s1": ctx_features["pos"][stack[-2]] if len(stack) >= 2 else "__EMPTY__",
        "b0": ctx_features["pos"][n0],
        "b1": ctx_features["pos"][n0 + 1] if len(words) > n0 + 1 else "__EMPTY__"
        }

    for x, y in [("s0", "b0"), ("s0", "s1"), ("s0", "b1")]:

        # gender/number/case agreements
        if morpho[x].get("number", "X") == morpho[y].get("number", "Y"):
            add("same_number_{}/{}", x, y)
        if morpho[x].get("gender", "X") == morpho[y].get("gender", "Y"):
            add("same_gender_{}/{}", x, y)
        if morpho[x].get("case", "X") == morpho[y].get("case", "X"):
            add("same_case_{}/{}", x, y)
        if morpho[x] == morpho[y]:
            add("same_morpho_s0/b0")

        # gender/number/case info
        add("case: {}={} {}={}", x, morpho[x].get("case", None), y, morpho[y].get("case", None))
        add("number: {}={} {}={}", x, morpho[x].get("number", None), y, morpho[y].get("number", None))
        add("gender: {}={} {}={}", x, morpho[x].get("gender", None), y, morpho[y].get("gender", None))

        # gender/number/case info conditionned on POS
        add("case: {}/{}={} {}/{}={}", x, pos[x], morpho[x].get("case", None), y, pos[y], morpho[y].get("case", None))
        add("number: {}/{}={} {}/{}={}", x, pos[x], morpho[x].get("number", None), y, pos[y], morpho[y].get("number", None))
        add("gender: {}/{}={} {}/{}={}", x, pos[x], morpho[x].get("gender", None), y, pos[y], morpho[y].get("gender", None))

    return features
