def dict_features(config, delexicalized=False, labeled=False, beam=False, morpho=False, embeddings=None):
    """Extract POS features for dependency parsing

    The feature templates of [Zhang and Nivre, 2011] are used.

    [Zhang and Nivre, 2011] Transition-based Dependency Parsing
    with Rich Non-local Features, ACL'11

    Parameters
    ----------
    - words, a list of strings
         the input sentence.
    - tags, a dictionnary of lists
         PoS corresponding to the words.
    - n0, an int
         the word we are extracting features for
    - stack
    - parse, a instance of `Parse`
         the partial parse tree
    - delexicalized, a boolean
         If true, o features describing words are considered so that a
         parser trained on one language can be applied on another
         language (cf. [McDonald et al, 2011]).

    """

    from .AbstractLocalParser import extract_features
    data = config.extract_features() if beam else extract_features(*config)

    (word, tag, labels, valencies, dist, hist_feat) = split_type(data)

    if delexicalized:
        word = [None for _ in word]

    cpos = [t.cpos if 'cpos' in dir(t) else t for t in tag]
    if morpho:
        lemma = [t.lemma if 'lemma' in dir(t) else "_" for t in tag]
        fpos = [t.fpos if 'fpos' in dir(t) else "_" for t in tag]
        ufeats = [t.ufeats if 'ufeats' in dir(t) else {} for t in tag]

    valencies = [min(x, 10) for x in valencies]
    # the distance is clipped to 5
    dist = min(dist, 5)

    features = {
        "word": list(word),
        "tag": list(cpos),
        "valencies": list(valencies),
        "distance": [dist],
        "history": list(hist_feat)
        }
    if labeled:
        features["labels"] = labels
    if morpho:
        features["lemma"] = lemma
        features["fpos"] = fpos
    return features


def zn11_features(config, delexicalized=False, labeled=False, beam=False, morpho=False, embeddings=None):
    """Extract POS features for dependency parsing

    The feature templates of [Zhang and Nivre, 2011] are used.

    [Zhang and Nivre, 2011] Transition-based Dependency Parsing
    with Rich Non-local Features, ACL'11

    Parameters
    ----------
    - words, a list of strings
         the input sentence.
    - tags, a dictionnary of lists
         PoS corresponding to the words.
    - n0, an int
         the word we are extracting features for
    - stack
    - parse, a instance of `Parse`
         the partial parse tree
    - delexicalized, a boolean
         If true, o features describing words are considered so that a
         parser trained on one language can be applied on another
         language (cf. [McDonald et al, 2011]).

    """

    from .AbstractLocalParser import extract_features
    data = config.extract_features() if beam else extract_features(*config)

    str_none = "None"
    data = [x if x is not None else str_none for x in data]

    (word, tag, labels, valencies, dist, hist_feat) = split_type(data)

    s0w, s1w, s2w, n0w, n1w, n2w, n0lw, n0l2w, n0rw, n0r2w, s0lw, s0l2w, s0rw, s0r2w, s0hw, s0h2w = word

    cpos = [t.cpos if 'cpos' in dir(t) else t for t in tag]
    s0p, s1p, s2p, n0p, n1p, n2p, n0lp, n0l2p, n0rp, n0r2p, s0lp, s0l2p, s0rp, s0r2p, s0hp, s0h2p = cpos

    n0lL, n0l2L, n0rL, n0r2L, s0lL, s0l2L, s0rL, s0r2L = labels

    valencies = [min(x, 10) for x in valencies]
    # the distance is clipped to 5
    dist = min(dist, 5)

    n0lv, n0rv, s0lv, s0rv = map(str, valencies)
    dist = str(dist)

    features = dict()

    def add(feat, lex=False, label=False, v=1):
        if delexicalized and lex:
            return
        if not labeled and label:
            return
        features[feat] = v

    # Baseline feature template from [Zhang and Nivre, 2011]
    # ======================================================

    # biais
    add("biais")

    # Single word features
    #
    # S0w/p
    add("s0w/p=" + s0w + "/" + s0p, True)
    # S0w
    add("s0w=" + s0w, True)
    # S0p
    add("s0p=" + s0p)
    # N0wp
    add("n0w/p=" + n0w + "/" + n0p, True)
    # N0w
    add("n0w=" + n0w, True)
    # N0p
    add("n0p=" + n0p)
    # N1wp
    add("n1w/p=" + n1w + "/" + n1p, True)
    # N1w
    add("n1w=" + n1w, True)
    # N1p
    add("n1p=" + n1p)
    # N2wp
    add("n2w/p=" + n2w + "/" + n2p, True)
    # N2w
    add("n2w=" + n2w, True)
    # N2p
    add("n2p=" + n2p)

    # Word pair features
    #
    # S0wpN0wp
    add("s0wp=" + s0w + "/" + s0p + "_and_n0wp=" + n0w + "/" + n0p, True)
    # S0wpN0w
    add("s0wp=" + s0w + "/" + s0p + "_and_n0w=" + n0w, True)
    # S0wN0wp
    add("s0w=" + s0w + "_and_n0wp=" + n0w + "/" + n0p, True)
    # S0wpN0p
    add("s0wp=" + s0w + "/" + s0p + "_and_n0p=" + n0p, True)
    # S0pN0wp
    add("s0p=" + s0p + "_and_n0wp=" + n0w + "/" + n0p, True)
    # S0wN0w
    add("s0w=" + s0w + "_and_n0w=" + n0w, True)
    # S0pN0p
    add("s0p=" + s0p + "_and_n0p=" + n0p)
    # N0pN1p
    add("n0p=" + n0p + "_and_n1p=" + n1p)

    # from three words
    #
    # N0pN1pN2p
    add("n0p=" + n0p + "_and_n1p=" + n1p + "_and_n2p=" + n2p)
    # S0pN0pN1p
    add("s0p=" + s0p + "_and_n0p=" + n0p + "_and_n1p=" + n1p)
    # S0hpS0pN0p
    add("s0hp=" + s0hp + "_and_s0p=" + s0p + "_and_n0p=" + n0p)
    # S0pS0lpN0p
    add("s0p=" + s0p + "_and_s0lp=" + s0lp + "_and_n0p=" + n0p)
    # S0pS0rpN0p
    add("s0p=" + s0p + "_and_s0rp=" + s0rp + "_and_n0p=" + n0p)
    # S0pN0pN0lp
    add("s0p=" + s0p + "_and_n0p=" + n0p + "_and_n0lp=" + n0lp)

    # New features templates
    # ======================
    #
    # distance
    # S0wd
    add("s0w/d=" + s0w + "/" + dist, True)
    # S0pd
    add("s0p/d=" + s0p + "/" + dist)
    # N0wd
    add("n0w/d=" + n0w + "/" + dist, True)
    # N0p=d
    add("n0pd=" + n0p + "/" + dist)
    # S0wN0wd
    add("s0wN0wd=" + s0w + "/" + n0w + "/" + dist, True)
    # S0pN0pd
    add("s0pN0pd=" + s0p + "/" + n0p + "/" + dist)

    # Valency
    # S0wv_r
    add("s0w=" + s0w + "_and_rv=" + s0rv, True)
    # S0pv_r
    add("s0p=" + s0p + "_and_rv=" + s0rv)
    # S0wv_l
    add("s0w=" + s0w + "_and_lv=" + s0lv, True)
    # S0pv_l
    add("s0p=" + s0p + "_and_lv=" + s0lv)
    # N0wv_l
    add("n0w=" + n0w + "_and_lv=" + n0lv, True)
    # N0pv_l
    add("n0p=" + n0p + "_and_lv=" + n0lv)

    # Unigrams
    # S0hw
    add("s0hw=" + s0hw, True)
    # S0hp
    add("s0hp=" + s0hp)
    # S0hL
#    add("s0hL=" + s0hL, label=True)
    # S0lw
    add("s0lw=" + s0lw, True)
    # S0lp
    add("s0lp=" + s0lp)
    # S0lL
    add("s0lL=" + s0lL, label=True)
    # S0rw
    add("s0rw=" + s0rw, True)
    # S0rp
    add("s0rp=" + s0rp)
    # S0rL
    add("s0rL=" + s0rL, label=True)
    # N0lw
    add("n0lw=" + n0lw, True)
    # N0lp
    add("n0lp=" + n0lp)
    # N0lL
    add("n0lL=" + n0lL, label=True)

    # Third-order features
    # S0h2w
    add("s0h2w=" + s0h2w, True)
    # s0h2p
    add("s0h2p=" + s0h2p)
    # s0h2L
#    add("s0h2L=" + s0h2L, label=True)
    # s0l2w
    add("s0l2w=" + s0l2w, True)
    # s0l2p
    add("s0l2p=" + s0l2p)
    # s0l2L
    add("s0l2L=" + s0l2L, label=True)
    # s0r2w
    add("s0r2w=" + s0r2w, True)
    # s0r2p
    add("s0r2p=" + s0r2p)
    # s0r2L
    add("s0r2L=" + s0r2L, label=True)
    # n0l2w
    add("n0l2w=" + n0l2w, True)
    # n0l2p
    add("n0l2p=" + n0l2p)
    # n0l2L
    add("n0l2L=" + n0l2L, label=True)
    # S0pS0lpS0l2p
    add("s0p=" + s0p + "_and_s0lp=" + s0lp + "_and_s0l2p=" + s0l2p)
    # S0pS0rpS0r2p
    add("s0p=" + s0p + "_and_s0rp=" + s0rp + "_and_s0r2p=" + s0r2p)
    # S0pS0hpS0h2p
    add("s0p=" + s0p + "_and_s0hp=" + s0hp + "_and_s0h2p=" + s0h2p)
    # N0pN0lpN0l2p
    add("n0p=" + n0p + "_and_n0lp=" + n0lp + "_and_n0l2p=" + n0l2p)

    # Label set
    # S0wS0Ll
#    add("s0w=" + s0w + "_and_Ll=" + s0Ll, True, label=True)
    # S0pS0Ll
#    add("s0p=" + s0p + "_and_Ll=" + s0Ll, label=True)
    # S0wS0Lr
#    add("s0w=" + s0w + "_and_Lr=" + s0Lr, True, label=True)
    # S0pS0Lr
#    add("s0p=" + s0p + "_and_Lr=" + s0Lr, label=True)
    # N0wN0Ll
#    add("n0w=" + n0w + "_and_Ll=" + n0Ll, True, label=True)
    # N0pN0Ll
#    add("n0p=" + n0p + "_and_Ll=" + n0Ll, label=True)

    # History of decisions
    for i, v in enumerate(hist_feat):
        add("hist@-" + str(i) + "=" + str(v))

    if morpho:
        #
        if 'ufeats' in dir(tag[0]) and type(tag[0].ufeats) == dict:
            for k,v in tag[0].ufeats.items():
                for v in v if type(v) else [v]:
                    add("s0f:"+k+"="+v)
        if 'ufeats' in dir(tag[3]) and type(tag[3].ufeats) == dict:
            for k,v in tag[3].ufeats.items():
                for v in v if type(v) else [v]:
                    add("n0f:"+k+"="+v)

    if embeddings:
        for i,v in enumerate(v for w in (s1w, s0w, n0w) for v in embeddings.get(w, embeddings["<UNK>"])):
            add(i, v=v)

    return features


def split_type(word_feats):
    (s0w, s1w, s2w,
     s0p, s1p, s2p,
     n0w, n1w, n2w,
     n0p, n1p, n2p,
     n0lv, n0lw, n0l2w, n0lp, n0l2p, n0lL, n0l2L,
     n0rv, n0rw, n0r2w, n0rp, n0r2p, n0rL, n0r2L,
     s0lv, s0lw, s0l2w, s0lp, s0l2p, s0lL, s0l2L,
     s0rv, s0rw, s0r2w, s0rp, s0r2p, s0rL, s0r2L,
     s0hw, s0hp,
     s0h2w, s0h2p,
     dist, hist_feat) = word_feats

    return ((s0w, s1w, s2w, n0w, n1w, n2w, n0lw, n0l2w, n0rw, n0r2w, s0lw, s0l2w, s0rw, s0r2w, s0hw, s0h2w),
            (s0p, s1p, s2p, n0p, n1p, n2p, n0lp, n0l2p, n0rp, n0r2p, s0lp, s0l2p, s0rp, s0r2p, s0hp, s0h2p),
            (n0lL, n0l2L, n0rL, n0r2L, s0lL, s0l2L, s0rL, s0r2L),
            (n0lv, n0rv, s0lv, s0rv),
            dist, hist_feat)
