from .AbstractStructuredPredictor import AbstractStructuredPredictor

import abc
import random, itertools
from operator import itemgetter

from structured_prediction.AbstractStructuredPredictor import Callback
from misc.itertools import next0, catch0

# algorithmic switches
no_final_ranking = False
p_reset = 0

EARLY_DET, EARLY_LONGEST_PREFIX, EARLY_TOP_SIBLING, EARLY_TOP_SIBLING_PAST, MAX_VIOLATION_DET, MAX_VIOLATION_FROM_INIT, MAX_VIOLATION_FROM_TOP_SIBLING, MAX_VIOLATION_ND, DYNAMIC_ALONG = range(9)

class AbstractBeamStructuredPredictor(AbstractStructuredPredictor):

    beam = 8

    @property
    def k(self): return self.beam

    def with_beam(self, beam):
        self.beam = beam
        return self

    @abc.abstractmethod
    def _initBeamConfig(self, sample, config=None): pass

################################################################################
##                                  BEAM SEARCH                               ##
################################################################################

    def forced_decoding(self, blackbox, state, depth=None, return_all=False, return_states=False, static=True):
        return self.lookahead(blackbox, state, time_limit=depth, return_all=return_all, return_states=return_states, static=static, forced_decoding=True)

    def lookahead(self, blackbox, state, check_cut_edge=True, check_single_reference=None, time_limit=None, remember_bad_beams=1, return_all=False, return_states=False, static=False, forced_decoding=False):
        """Performs beam search decoding from a given parse configuration up to next (state-level) error, then yields every (erroneous) state up to final configurations

        This function controls beam decoding, oracle computation and how errors are defined.

        Decoding is lazy, which means that:
            - computations and side-effects do not happen until the first state is checked,
            - and when used in early update mode, decoding stops after the first error.

        Beam search decoding follows [Zhang and Clark, 2008] and state-level errors are flagged when all hypotheses are erroneous, according to [Aufrant et al, 2016].

        [Zhang and Clark, 2008] A Tale of Two Parsers: Investigating and Combining Graph-based and Transition-based Dependency Parsing Using Beam-search, EMNLP'08
        [Aufrant et al, 2017] Don’t Stop Me Now! Using Global Dynamic Oracles to Correct Training Biases of Transition-Based Dependency Parsers, EACL'17

        Arguments
        ---------
        - parse: an instance of AbstractBeamParse
            The initial parse configuration.
            Its past errors are taken for granted, and all subsequent configurations derive from `parse`.

        Keyword arguments
        -----------------
        - check_cut_edge: a boolean (default: True)
        - check_single_reference: an instance of AbstractBeamParse, or None (default: None)
        - time_limit: an int, or None (default: None)
        - remember_bad_beams: a float (default: 1)
            cf doc

        Global controls
        ---------------
        - p_reset: a float

        Returns
        -------
        - a generator yielding states as `(kbest, (successors, guilty_state))`:
            - kbest: a BeamState list
                The `self.k`-best derivation hypotheses in this beam state, in decreasing order of score.
            - successors: a list of (float, AbstractBeamParse, int) tuples
                The list of derivation hypotheses just before last k-truncation, and with last action not executed.
                The tuple contains the future score, the configuration on which to apply the given move, and the move to execute.
                For the first yielded state, it necessarily contains a correct derivation.
            - guilty_state: a (list of AbstractBeamParses, list of tuples) tuple
                The `kbest` and `successors` values at the step where the first `remember_bad_beams` (in %) part of the beam was in error.
                For a `remember_bad_beams` value of 1, it is not computed (value would be the same as `kbest` and `successors`) and defaults to ([], None).
            States are yielded in increasing order of depth (from `parse` to final configurations).
            The generator yields at least one value: if no error is found, the final state is yielded.
            If the initial state is already final, it is the only value yielded.

        Side-effects
        ------------
        - parse.alive is overwritten to True
        - parse.configuration is overwritten
        """

        # list of hypotheses
        kbest = state if type(state)==list else [state]
        successors = []
        guilty_state = (None, [])

        if not forced_decoding:
            # let's assume our past errors and do our best with them
            for config in kbest: config.alive = True
            error_mode = False

        if check_single_reference:
            reverse_golden_path = list(check_single_reference.predecessors)
            golden_path = reversed(reverse_golden_path[:-kbest[0].depth] if kbest[0].depth else reverse_golden_path)

        timeline = itertools.count() if time_limit is None else range(time_limit)

        for _ in timeline:

            if not forced_decoding:
                # force suboptimal training
                if random.random() < p_reset and not error_mode and not kbest[0].final:
                    kbest = kbest[:1]
                    kbest[0].alive = True

            if not forced_decoding:
                if remember_bad_beams < 1:
                    import math
                    if any(config.alive for config in kbest[:math.ceil(self.k * remember_bad_beams)]):
                        guilty_state = (None, [])
                    elif guilty_state[0] is None:
                        guilty_state = (kbest, successors)

            # the best hypothesis has reached the end of derivation, so have the others
            if kbest[0].final and kbest[0].constant_length or all(hyp.final for hyp in kbest):
                break

            if forced_decoding:
                if return_states:
                    yield False, kbest if return_all else kbest[0]
            else:
                # gold out of beam
                if not any(config.alive for config in kbest):
                    error_mode = True

                if error_mode: yield False, (kbest, (successors, guilty_state))

            # successors of current kbest
            successors = []
            kbest = ((p, self.features(p, beam=True), p.valid_moves()) for p in kbest)
            for config, features, valid_moves in kbest:
                constrained_valid_moves = list(config.min_cost(valid_moves, constraints=True))
                gold_moves = config.min_cost(constrained_valid_moves) if config.gold else constrained_valid_moves
                moves = gold_moves if forced_decoding else constrained_valid_moves

                evaluated_moves = [] if self.k == 1 and len(moves) == 1 else moves ###

                yield True, self._joint_wrap(1, features, tuple(evaluated_moves))
                _X, (_, scores) = next(blackbox.next)
                if not scores: scores = {move: 0 for move in moves}
                config.configuration = _X, gold_moves, scores

                extra = yield from self._joint_lookahead(blackbox, features)

                moves = constrained_valid_moves if not forced_decoding else gold_moves
                assert moves
                successors.extend((config.score + scores[move], config, move, extra) for move in moves)
            successors.sort(key=itemgetter(2 if static else 0), reverse=True)

            # advance to next step
            future_kbest = successors[:(1 if static else self.k)]
            kbest = [config.transition(move, score - config.score, **extra) for score, config, move, extra in future_kbest]

            # synchronously advance the reference
            if check_single_reference: single_reference = next(golden_path)

            if not forced_decoding:
                # check life status
                for config in kbest:
                    if check_single_reference:
                        assert single_reference.depth == config.depth
                    config.alive = (config.previous.alive
                                   and (not check_cut_edge or config.last_move in config.previous.configuration[1])
                                   and (not check_single_reference or config.last_move == single_reference.last_move))

        if forced_decoding:
            yield False, kbest if return_all else kbest[0]
        else:
            yield False, (kbest, (successors, guilty_state))

################################################################################
##                                 ERROR STATES                               ##
################################################################################

    def _infer(self, blackbox, sample, forced_decoding=False,
               oracle=None, exploration_rate=1, check_single_reference=False, beam_update=False,
               feature_file=None, cb=Callback(), **kwargs):
        """Returns a series of erroneous (or final, if not possible) states with the corresponding reference, taking each previous error as given.

        This function controls the Parse object initialization, the reference choice and the training strategy. It can be used both for training and predicting.

        Decoding is lazy, which means that:
            - computations and side-effects do not happen until the first state is checked,
            - and when used in early update mode, decoding stops after the first error.

        Arguments
        ---------
        - sample: a Pan4 tuple with bottom-up representations
            The tokenized input sentence, or training example.
            The training mode is enabled when the first or unique `heads` information is not None: in that case it is used as the gold parse tree. It can be partial, but in this case the given forest must be compatible with a full projective tree.
            If (partial) constraints are specified, the search spece (at train and prediction time) if restrained to trees compatible with the not-None heads.

        Keyword arguments
        -----------------
        - oracle: an int, or None
        - exploration_rate: a float (default: 1)
        - beam_update: a boolean (default: False)
        - check_cut_edge: a boolean (default: True)
        - check_single_reference: a boolean (default: False)
        - time_limit: an int, or None (default: None)
        - remember_bad_beams: a float (default: 1)
        - feature_file: a filehandler, or None (default: None)
            cf doc

        Global controls
        ---------------
        - no_final_ranking: a boolean
        - p_reset: a float

        Returns
        -------
        - a generator yielding update states as `(kbest, reference)`:
            - kbest: a BeamState list
                The `self.k`-best derivation hypotheses in this beam state, in decreasing order of score.
            - reference: an AbstractBeamParse instance
                In prediction mode and when no error has been found, it is a dummy reference equal to the prediction.
            Update states are yielded in increasing order of depth (from `parse` to final configurations).
            The generator yields at least one value: if no error is found, the final state is yielded.
            A final state is always yielded. #

        Side-effects
        ------------
        - increments `regular_stops` with the number of encountered final configurations (excluding configurations only encountered during oracle computation)
        - features are dumped for each evaluated configuration, if a `feature_file` has been registered
        """

        config = self._initBeamConfig(sample)

        if forced_decoding:
            depth = None
            if type(forced_decoding) == tuple: forced_decoding, depth = forced_decoding
            if forced_decoding != True:
                config = self._initBeamConfig(sample, config=forced_decoding)

            kbest = yield from next0(self.forced_decoding(blackbox, config, depth=depth, return_all=True))
            yield False, (kbest, None)
            return

        precomputed = yield from self.oracle_precomputation(blackbox, oracle, config, feature_file)
        if check_single_reference and precomputed: check_single_reference = precomputed

        regular_stop = False
        while not regular_stop:
            decode = self.lookahead(blackbox, config, check_single_reference=check_single_reference, **kwargs)
            rank = -1 if beam_update else 0
            candidates = yield from self.lookfurther(blackbox, oracle, decode, config, rank, precomputed, feature_file)

            depths = reversed(list(zip(*candidates)))
            def gap(state):
                plus = state[0]
                minus = state[1][0][rank]
                return minus.score - plus.score

            state = max(depths, key=gap)
            prediction, reference, kbest = yield from self.refchoice(blackbox, oracle, state, rank, precomputed)
            regular_stop = prediction.final

            cb.on_update_infer(regular_stop)
            if no_final_ranking and gold and any(config.alive for config in kbest):
                break
            yield False, (kbest, reference)

            config = prediction if random.random() < exploration_rate else reference

################################################################################
##                                    ORACLES                                 ##
################################################################################

    def oracle_precomputation(self, blackbox, oracle, config, feature_file):
        if oracle in [EARLY_DET, MAX_VIOLATION_DET]:
            self.model.dump_features(None)
            static_reference = yield from next0(self.forced_decoding(blackbox, config, static=True))
            self.model.dump_features(feature_file)
            return static_reference

    def lookfurther(self, blackbox, oracle, decode, init_config, rank, precomputed, feature_file):
        if oracle not in [MAX_VIOLATION_FROM_INIT, MAX_VIOLATION_DET, MAX_VIOLATION_FROM_TOP_SIBLING, MAX_VIOLATION_ND]:
            negative = [(yield from next0(decode))]
            # dummy reference: prediction
            positive = [kbest[0 if kbest[0].final else rank] for kbest, _ in negative]
            return positive, negative

        negative = list((yield from catch0(decode)))
        self.model.dump_features(None)
        _, (successors, _) = negative[0]

        if oracle in [MAX_VIOLATION_DET, MAX_VIOLATION_FROM_INIT, MAX_VIOLATION_FROM_TOP_SIBLING]:
            if oracle == MAX_VIOLATION_DET:
                config, depth = precomputed, 0
            elif oracle == MAX_VIOLATION_FROM_INIT:
                config, depth = init_config, None
            elif oracle == MAX_VIOLATION_FROM_TOP_SIBLING:
                config, depth = next(self.alive_siblings(successors)), None
            reference = yield from next0(self.forced_decoding(blackbox, config, depth=depth))
            positive = reference.predecessors
        elif oracle == MAX_VIOLATION_ND:
            references = yield from catch0(self.forced_decoding(blackbox, list(self.alive_siblings(successors)), return_states=True))
            positive = reversed(list(references))

        self.model.dump_features(feature_file)
        return positive, reversed(negative)

    def refchoice(self, blackbox, oracle, state, rank, precomputed):
        reference, (kbest, history) = state
        prediction = kbest[0 if kbest[0].final else rank]

        if oracle == EARLY_TOP_SIBLING_PAST:
            kbest, successors = history[1]
            state = None, (kbest, (successors, None))
            return self.refchoice(blackbox, EARLY_TOP_SIBLING, state, rank, precomputed)

        if oracle == EARLY_DET:
            reference = self.same_depth_ancestor(precomputed, prediction)
        elif oracle == EARLY_TOP_SIBLING:
            reference = next(self.alive_siblings(history[0]))
        elif oracle == EARLY_LONGEST_PREFIX:
            reference = yield from next0(self.closest_alive_siblings(blackbox, prediction))

        return prediction, reference, kbest

    def alive_siblings(self, successors):
        alive_parents = (succ for succ in successors if succ[1].alive)
        return (parent.transition(move, parent.configuration[2][move], **extra) for _,parent,move,extra in alive_parents if move in parent.configuration[1])
    def closest_alive_siblings(self, blackbox, prediction):
        ancestor = next(ancestor for ancestor in prediction.predecessors if ancestor.alive)
        return self.forced_decoding(blackbox, ancestor, prediction.depth - ancestor.depth)
    def same_depth_ancestor(self, longest, shortest):
        return next(state for state in longest.predecessors if state.depth == shortest.depth)
