from misc.io import Listener

import abc

class Callback:
    def on_train_begin(self, model, dataset): pass
    def on_train_end(self, model, *args, **kwargs): pass
    def on_epoch_begin(self, model, itn): pass
    def on_epoch_end(self, model, itn): pass
    def on_sample_begin(self, model, sample): pass
    def on_sample_end(self, model, sample, before, did_update, final_config): pass

    shuffle = True

    epoch_eval = None
    def before_epoch_val(self, model, itn, *args, **kwargs): pass
    def after_epoch_val(self, model, itn, perf, *args, **kwargs): pass

    polish = True
    def on_polish(self, *args, **kwargs): pass

    def on_timestep(self, suboptimal, config, XXX): pass
    def on_update_begin_learn(self, suboptimal, y, _y, sample): pass
    def on_update_end_learn(self, suboptimal, config): pass
    def on_update_infer(self, full_search, *args, **kwargs): pass

class EndOfEpoch: pass
EndOfEpoch = EndOfEpoch()

class AbstractStructuredPredictor(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def _defaultmodel(self, *args, **kwargs):
        pass

    def _initmodel(self, *args, **kwargs):
        self.model = self._defaultmodel(*args, **kwargs)

    def x_to_modelspace(self,x): return self.model.x_to_modelspace(x)
    def y_to_modelspace(self,y): return self.model.y_to_modelspace(y)

    #
    def update(self, *config):
        Y, _Y = map(self.y_to_modelspace, config[:2])
        return self.model.update(Y, _Y, *config[2:])

    def train_on_batch(self, *batch):
        Ys, _Ys = map(self.ys_to_modelspace, batch[:2])
        batch = list(map(list, batch[2:])) #
        return self.model.train_on_batch(Ys, _Ys, *batch)

    def xs_to_modelspace(self,xs): return list(map(self.x_to_modelspace, xs))
    def ys_to_modelspace(self,ys): return list(map(self.y_to_modelspace, ys))

    constrained_predictions = False

    def predict(self, _x, classes=None):
        if self.constrained_predictions: _x, classes = _x
        _X = self.x_to_modelspace(_x)
        pred = self.model.predict(_X, classes, scores=True)
        return _X, pred
    def predict_on_batch(self, _xs, classes=None):
        if self.constrained_predictions:_xs, classes = map(list, zip(*_xs))
        _Xs = self.xs_to_modelspace(_xs)
        preds = self.model.predict_on_batch(_Xs, classes, scores=True)
        return list(zip(_Xs, preds))

    def gen_dyn_config(self, _xys, batched=True, cb=Callback()):
        from itertools import starmap
        if batched:
            from misc.itertools import unzip_justify_compress
            feedback, outputs, futures = _xys
            predictions = starmap(lambda b, mask: (self.predict_on_batch(b), mask), feedback)
            predictions = unzip_justify_compress(predictions, batched=len(futures))
            for future, it in zip(futures, predictions):
                future.next = it
        else:
            blackbox = type('', (), {})()
            feedback, outputs = _xys(blackbox)
            blackbox.next = map(self.predict, feedback)
        return outputs

    def gen_predictions(self, _xs, batched=True):
        from itertools import starmap
        if batched:
            from misc.itertools import unzip_justify_compress
            feedback, outputs, futures = _xs
            predictions = starmap(lambda b, mask: (self.predict_on_batch(b), mask), feedback)
            predictions = unzip_justify_compress(predictions, batched=len(futures))
            for future, it in zip(futures, predictions):
                future.next = it
        else:
            blackbox = type('', (), {})()
            feedback, outputs = _xs(blackbox)
            blackbox.next = map(self.predict, feedback)
        return outputs

    @abc.abstractmethod
    def xs(self, sample, cb=Callback(), **kwargs):
        pass

    @abc.abstractmethod
    def xys(self, sample, cb=Callback(), **kwargs):
        pass

    def _xs(self, sample, cb=Callback(), **kwargs):
        from misc.itertools import deinterleave
        return lambda blackbox: deinterleave(self.xs(sample, cb=cb, **kwargs)(blackbox))

    def _xys(self, sample, cb=Callback(), **kwargs):
        from functools import partial
        f_xys = partial(self.xys, cb=cb, **kwargs)
        from misc.itertools import deinterleave
        return lambda blackbox: deinterleave(f_xys(sample[0], **(sample[1]))(blackbox))

    def _batchxs(self, batch, cb=Callback(), **kwargs):
        from misc.itertools import zip_justify_compress
        futures = [type('', (), {})() for _ in batch]
        feedback, out = zip(*[self._xs(sample, cb=cb, **kwargs)(bb) for sample, bb in zip(batch, futures)])
        return zip_justify_compress(*feedback), (out), futures

    def _batchxys(self, batch, cb=Callback(), **kwargs):
        from misc.itertools import zip_justify_compress, zip_justify
        futures = [type('', (), {})() for _ in batch]
        feedback, out = zip(*[self._xys(sample, cb=cb, **kwargs)(bb) for sample, bb in zip(batch, futures)])
        return zip_justify_compress(*feedback), zip_justify(*out), futures

    def train_online_configs(self, sample, batched=True, cb=Callback(), **kwargs):
        from functools import partial
        _xys = (self._batchxys if batched else self._xys)(sample, cb=cb, **kwargs)
        return self.gen_dyn_config(_xys, batched=batched, cb=cb)

    def train_online(self, dataset, batched=1, cb=Callback(), **kwargs):
        from functools import partial
        process = partial(self.train_online_configs, batched=bool(batched), cb=cb, **kwargs)
        # EndOfEpoch values can delimit epochs
        from itertools import takewhile
        dataset = takewhile(lambda x: x is not EndOfEpoch, dataset)

        # if not batched: 1-sized batches
        from misc.itertools import batchmap, multiplex
        for batch in batchmap(dataset, process, batched) if batched else multiplex(dataset, process, 1):
            self.train_on_batch(*zip(*batch))

    def polish(self, outputs):
        return outputs

    def process_sample(self, sample, batched=False, **kwargs):
        _xs = (self._batchxs if batched else self._xs)(sample, **kwargs)
        return self.gen_predictions(_xs, batched=batched)

    def process_dataset(self, dataset, batched=1, **kwargs):
        from functools import partial
        process = partial(self.process_sample, batched=bool(batched), **kwargs)
        if batched:
            from misc.itertools import batchmap
            return batchmap(dataset, process, batched)
        return map(process,dataset)

    def process(self, dataset, preprocess=lambda x:x, postprocess=lambda *x:x, batched=1, raw=False, **kwargs):
        from itertools import tee, starmap
        dataset, inputs = tee(dataset)
        inputs = (preprocess(ex) for ex in inputs)
        outputs = self.process_dataset(inputs, batched=batched, **kwargs)
        if not raw: outputs = map(self.polish, outputs)
        return starmap(postprocess, zip(dataset, outputs))

    def strategy(self, strategy, n_epoch):
        return {}

    def epoch_strategy(self, strategy, itn, n_epoch):
        return {}

    def sample_strategy(self, strategy, sample, itn, n_epoch):
        return {}

    def train(self, dataset, n_epoch, strategy=0, listener=Listener, feature_file=None, crossvalidate=False, preprocess=lambda x:x, reshape=lambda x,itn:x, cb=Callback(), batched=1):
        #
        dataset = map(preprocess, dataset)
        if cb.shuffle:
            dataset = list(dataset)

        cb.on_train_begin(self, dataset)

        model_states = []

        #
        self.model.dump_features(feature_file)

        from functools import partial
        train = partial(self.train_online, batched=bool(batched), cb=cb, **(self.strategy(strategy, n_epoch)))

        import itertools #
        for itn in itertools.count() if crossvalidate else range(n_epoch):
            cb.on_epoch_begin(self, itn)

            batch_size = batched if not batched or type(batched) == int else batched(itn)
            self.model.gradient_normalization = batch_size if batch_size else 1
            epoch_train = partial(train, batched=batch_size, **(self.epoch_strategy(strategy, itn, n_epoch)))

            if cb.shuffle:
                import random
                random.shuffle(dataset)
            itn_dataset = (reshape(sample, itn) for sample in listener.iter(dataset))
            itn_dataset = (sample for sample in itn_dataset if sample)
            itn_dataset = ((sample, self.sample_strategy(strategy, sample, itn, n_epoch)) for sample in itn_dataset)

            epoch_train(itn_dataset, cb=cb)

            del self.model.gradient_normalization
            cb.on_epoch_end(self, itn)

            if cb.epoch_eval:
                self.model.dump_features(None)

                cb.before_epoch_eval(itn)
                perf = self.model.with_polished(lambda: cb.epoch_eval(self))
                cb.after_epoch_eval(self, itn, perf)

                if crossvalidate:
                    if type(crossvalidate) not in [int,float] or crossvalidate < 1:
                        crossvalidate = 1
                    # crossvalidate = n: stop after n score drops ==> remember n models
                    # crossvalidate = n.5: stop after n score drops followed by a drop or small increase (ie smaller than the last drop) ==> remember n+1 models
                    decreasing = len(model_states) >= crossvalidate and all(old[-1] >= new[-1] for old, new in zip(model_states, model_states[1:]))

                    if decreasing and model_states[-1 if int(crossvalidate) == crossvalidate else -2][-1] >= perf and itn + 1 >= n_epoch + crossvalidate:
                        self.model.restore(model_states[0][:-1])
                        break
                    else:
                        if len(model_states) >= crossvalidate:
                            model_states = model_states[1:]
                        model_states.append(self.model.backup() + (perf,))

                self.model.dump_features(feature_file)

            if feature_file is not None:
                import pickle
                pickle.dump({}, feature_file)

        if cb.polish:
            cb.on_polish(self)
            self.model.polish()

        cb.on_train_end(self)

def display_routine(testset, keep, display):
    for sample in testset:
        if keep(sample): display(sample)

def contains_type(by_type, bt):
    if by_type == bt: return True
    if type(by_type) in [tuple,list]:
        for sub_type in by_type:
            if contains_type(sub_type, bt): return True
    if type(by_type) == dict:
        for sub_type in by_type.values():
            if contains_type(sub_type, bt): return True
    return False

def eval_routine(testset, localzip, config, keep, category, eval_type=None, dump_errors=None):
    from collections import defaultdict
    corrects, totals = defaultdict(lambda: defaultdict(int)), defaultdict(lambda: defaultdict(int))

    if dump_errors:
        errors = defaultdict(lambda: defaultdict(int))

    for sample in testset:
        from functools import partial
        _keep = partial(keep, ex=sample)
        for tok in filter(_keep, localzip(sample)):
            _x, _y, y = config(tok)
            for k,sub in (eval_type if type(eval_type)==dict else {None:eval_type}).items():
                cat = category(tok, sample, sub)
                if y == _y: corrects[k][cat] += 1
                totals[k][cat] += 1

            if dump_errors:
                errors[_x][_y,y] += 1

    if dump_errors:
        import pickle
        with open(dump_errors, "ab") as fh:
            pickle.dump(errors, fh, pickle.HIGHEST_PROTOCOL)

    scores = defaultdict(lambda: defaultdict(lambda: (0.,0)))
    for sub in totals:
        for cat in totals[sub]:
            scores[sub][cat] = (corrects[sub][cat]/totals[sub][cat], totals[sub][cat])

    if type(eval_type) != dict:
        scores = scores[None]
    if eval_type is None:
        return scores[None][0]
    return scores

def matrix_routine(testset, localzip, config, keep, category, eval_type=None):
    from collections import defaultdict
    matrix = defaultdict(lambda: defaultdict(int))

    for sample in testset:
        from functools import partial
        _keep = partial(keep, ex=sample)
        for tok in filter(_keep, localzip(sample)):
            for k,sub in (eval_type if type(eval_type)==dict else {None:eval_type}).items():
                if type(sub) == tuple:
                    gold_sub, pred_sub = sub
                else:
                    gold_sub = pred_sub = sub
                gold_cat = category(tok, sample, gold_sub, gold=True)
                pred_cat = category(tok, sample, pred_sub, gold=False)
                matrix[k][gold_cat,pred_cat] += 1

    if type(eval_type) != dict:
        matrix = matrix[None]
    if eval_type is None:
        return matrix[None,None]
    return matrix

def stats_routine(dataset, localzip, config, keep, category, eval_type=None):
    from collections import defaultdict
    matrix = defaultdict(lambda: defaultdict(int))

    for sample in dataset:
        from functools import partial
        _keep = partial(keep, ex=sample)
        for tok in filter(_keep, localzip(sample)):
            for k,sub in (eval_type if type(eval_type)==dict else {None:eval_type}).items():
                if type(sub) == tuple:
                    coarse_sub, fine_sub = sub
                else:
                    coarse_sub = fine_sub = sub
                coarse_cat = category(tok, sample, coarse_sub, gold=True)
                fine_cat = category(tok, sample, fine_sub, gold=False)
                matrix[k][coarse_cat,fine_cat] += 1

    if type(eval_type) != dict:
        matrix = matrix[None]
    if eval_type is None:
        return matrix[None,None]
    return matrix

def compare_routine(testset, localzip, config, keep, category, eval_type=None):
    from collections import defaultdict
    corrects, totals = ([defaultdict(lambda: defaultdict(int)) for _ in range(2)] for _ in range(2))
    join_corrects, join_same, join_totals = (defaultdict(lambda: defaultdict(int)) for _ in range(3))

    from itertools import tee
    testset, testsets = zip(*map(tee, testset))

    for i in range(2):
        for sample in testsets[i]:
            from functools import partial
            _keep = partial(keep, ex=sample)
            for tok in filter(_keep, localzip(sample)):
                _x, _y, y = config(tok)
                for k,sub in (eval_type if type(eval_type)==dict else {None:eval_type}).items():
                    cat = category(tok, sample, sub)
                    if y == _y: corrects[i][k][cat] += 1
                    totals[i][k][cat] += 1

    #
    for sample in zip(*testset):
        from functools import partial
        _keep1 = partial(keep, ex=sample[0])
        _keep2 = partial(keep, ex=sample[1])
        for tok in filter(lambda x: _keep1(x[0]) and _keep2(x[1]), zip(*map(localzip, sample))):
            _x1, _y1, y = config(tok[0])
            _x2, _y2, _ = config(tok[1])
            for k,sub in (eval_type if type(eval_type)==dict else {None:eval_type}).items():
                cat = category(tok[0], sample[0], sub)
                if y == _y1 == _y2: join_corrects[k][cat] += 1
                if _y1 == _y2: join_same[k][cat] += 1
                join_totals[k][cat] += 1

    #
    reproduce = defaultdict(lambda: defaultdict(float))
    relative_scores = defaultdict(lambda: defaultdict(float))
    agreement = defaultdict(lambda: defaultdict(float))
    for sub in corrects[0]:
        for cat in corrects[0][sub]:
            reproduce[sub][cat] = (join_corrects[sub][cat]/corrects[0][sub][cat], corrects[0][sub][cat])
    for sub in totals[1]:
        for cat in totals[1][sub]:
            relative_scores[sub][cat] = ((corrects[1][sub][cat]/totals[1][sub][cat])/(corrects[0][sub][cat]/totals[0][sub][cat]) if corrects[0][sub][cat] else float('inf'), totals[1][sub][cat])
    for sub in join_totals:
        for cat in join_totals[sub]:
            agreement[sub][cat] = (join_same[sub][cat]/join_totals[sub][cat], join_totals[sub][cat])

    if type(eval_type) != dict:
        reproduce = reproduce[None]
        relative_scores = relative_scores[None]
        agreement = agreement[None]
    if eval_type is None:
        reproduce = reproduce[None][0]
        relative_scores = relative_scores[None][0]
        agreement = agreement[None][0]
    return reproduce, relative_scores, agreement

def matrix_compare_routine(testset, localzip, config, keep, category, eval_type=None):
    from collections import defaultdict
    matrix = [defaultdict(lambda: defaultdict(int)) for _ in range(2)]

    for i in range(2):
        for sample in testset[i]:
            from functools import partial
            _keep = partial(keep, ex=sample)
            for tok in filter(_keep, localzip(sample)):
                for k,sub in (eval_type if type(eval_type)==dict else {None:eval_type}).items():
                    if type(sub) == tuple:
                        gold_sub, pred_sub = sub
                    else:
                        gold_sub = pred_sub = sub
                    gold_cat = category(tok, sample, gold_sub, gold=True)
                    pred_cat = category(tok, sample, pred_sub, gold=False)
                    matrix[i][k][gold_cat,pred_cat] += 1

    relative_matrix = defaultdict(lambda: defaultdict(int))
    #
    for sub in matrix[0]:
        for cat in matrix[0][sub]:
            relative_matrix[sub][cat] = (matrix[1][sub][cat]/matrix[0][sub][cat], matrix[0][sub][cat])
    for sub in matrix[1]:
        for cat in matrix[1][sub]:
            relative_matrix[sub][cat] = (matrix[1][sub][cat]/matrix[0][sub][cat] if matrix[0][sub][cat] else float('inf'), matrix[0][sub][cat] + matrix[1][sub][cat])

    if type(eval_type) != dict:
        relative_matrix = relative_matrix[None]
    if eval_type is None:
        return relative_matrix[None,None]
    return relative_matrix
