from misc.io import Listener

import abc
from collections import defaultdict
from operator import itemgetter

from structured_prediction.AbstractStructuredPredictor import AbstractStructuredPredictor, Callback, eval_routine

START = ['-START2-', '-START-']
END = ['-END-', '-END2-']

def spelling_pattern(w):
    sp = []
    for c in w:
        if c != c.lower():
            if not sp or sp[-1]!='X':
                sp.append('X')
        elif c != c.upper():
            if not sp or sp[-1]!='x':
                sp.append('x')
        else:
            sp += c
    return "".join(sp)

class AbstractPOSTagger(AbstractStructuredPredictor):

    '''Greedy Averaged Perceptron POS tagger'''

    def __init__(self):
        self.tagdict = {}
        self.classes = set()

    #
    def initialize(self, sentences):
        self._make_tagdict(sentences)
        self._initmodel(sentences)

    def _make_tagdict(self, sentences):
        '''Make a tag dictionary for single-tag words.'''
        counts = defaultdict(lambda: defaultdict(int))
        for sent in sentences:
            for word, pos_tag in zip(sent[0], sent[1]):
                counts[word][pos_tag] += 1
                self.classes.add(pos_tag)

        freq_thresh = 20
        ambiguity_thresh = 0.97
        for word, tag_freqs in counts.items():
            tag, mode = max(tag_freqs.items(), key=itemgetter(1))
            n = sum(tag_freqs.values())
            # Don't add rare words to the tag dictionary
            # Only add quite unambiguous words
            if n >= freq_thresh and (float(mode) / n) >= ambiguity_thresh:
                self.tagdict[word] = tag

    def xs(self, sample, cb=Callback(), **kwargs):
        words = sample
        context = START + [self._normalize(w) for w in words] + END
        def xs(blackbox):
            prev2, prev = START
            for i, word in enumerate(words):
                tag = self.tagdict.get(word)
                if not tag:
                    yield True, self._get_features(len(START) + i, word, context, prev, prev2)
                    _, (tag, _) = next(blackbox.next)
                yield False, tag
                prev2, prev = prev, tag

        return xs

    def xys(self, sample, cb=Callback(), **kwargs):
        words, tags = sample
        f_next = self.f_next(**kwargs)
        context = START + [self._normalize(w) for w in words] + END
        def xys(blackbox):
            prev2, prev = START
            for i, word in enumerate(words):
                feats = self._get_features(len(START) + i, word, context, prev, prev2)
                yield True, feats
                _X, (_y, _scores) = next(blackbox.next)
                cb.on_timestep(42, (tags[i], _y, _X, _X, 42), 42)
                yield False, (tags[i], _y, _X)
                prev2, prev = prev, f_next(feats, tags[i], _y)

        return xys

    def f_next(self, exploration_rate=1):
        from random import random
        return lambda x,y,_y: _y if random() < exploration_rate else y

    def polish(self, outputs):
        return list(outputs)

    def tag(self, *args, **kwargs):
        return list(self.process_sample(*args, batched=False, **kwargs))

    def _normalize(self, word):
        if '-' in word and word[0] != '-':
            return '!HYPHEN'
        elif word.isdigit() and len(word) == 4:
            return '!YEAR'
        elif word[0].isdigit():
            return '!DIGITS'
        else:
            return word.lower()

    @abc.abstractmethod
    def _get_features(self, i, word, context, prev, prev2):
        pass

    def epoch_strategy(self, strategy, itn, n_epoch):
        return strategies(self, strategy, itn, n_epoch)

def train_pos_tagger(tagger, sentences, n_epoch, strategy=0, listener=Listener, iter_callback=None, average=True, feature_file=None, crossvalidate=False):

    # suppose either no padding or ROOT token at the end
    preprocess = lambda ex: (ex[0][1:-1],ex[1][1:-1]) if len(ex)>2 else ex

    from threading import Lock
    logger = type('', (), {'lock': Lock()})()

    def on_train_begin(tagger, dataset):
        ###
        listener.msg("collect word frequence information for feature extraction")

        tagger.initialize(dataset)
        tagger.curve = []

    def on_epoch_begin(tagger, itn):
        listener.msg("start epoch n°{}".format(itn + 1))
        logger.correct = logger.total = 0

    def after_epoch_eval(tagger, itn, perf):
        if type(perf) in [ int , float ]:
            listener.msg("\nScore at iteration {}: {:.3%}".format(itn + 1, perf))
        tagger.curve.append((itn + 1, perf))

    cb = Callback()

    cb.on_train_begin = on_train_begin
    cb.on_epoch_begin = on_epoch_begin
    cb.on_epoch_end = lambda tagger, itn: listener.msg(
        "\nerror rate (train) at iteration {}: {:.3%}".format(itn + 1, (logger.total - logger.correct) / logger.total))

    cb.before_epoch_eval = lambda *_: listener.msg("Callback...")
    cb.epoch_eval = iter_callback
    cb.after_epoch_eval = after_epoch_eval

    cb.on_polish = lambda *_: listener.msg('Averaging weights')

    def on_timestep(suboptimal, config, XXX):
        y, _y = config[:2]
        correct = y==_y
        with logger.lock:
            if correct: logger.correct += 1
            logger.total += 1
    cb.on_timestep = on_timestep
    cb.polish = average

    if crossvalidate: crossvalidate = 2

    tagger.train(sentences, n_epoch, strategy=strategy, listener=listener, feature_file=feature_file, crossvalidate=crossvalidate,
        preprocess=preprocess, cb=cb, batched=lambda itn: 2**itn)

    return tagger

def process_pos_tagger(tagger, dataset, listener=Listener, test=False):
    dataset = listener.iter(dataset)
    if not tagger: return dataset

    preprocess = lambda ex: ex[0] if len(ex)==2 else ex[0][1:-1]

    from pan4.format import pad_tokens
    def postprocess(ex, tags):
        if len(ex)!=2: tags = pad_tokens(tags)
        gold_tags = ex[1]
        if test:
            tags = (gold_tags[:-1] if type(gold_tags) == tuple else (gold_tags,)) + (tags,)
        return ex[:1] + (tags,) + ex[2:]

    return tagger.process(dataset, preprocess=preprocess, postprocess=postprocess, batched=1000)

WORD, POS = 1, 2

def eval_read(testset):

    def read_test(gold, pred):
        from conll.io import rt, feat2tag
        with open(gold, "rt", encoding="utf-8", errors="replace") as gold, open(pred, "rt", encoding="utf-8", errors="replace") as pred:
            gold = feat2tag(rt(gold, multiroots=True))
            pred = feat2tag(rt(pred, multiroots=True))
            for g,p in zip(gold,pred):
                while g[0] != p[0]: g = next(gold)
                (w,gt,h,l),(_,t,_,_) = g,p
                yield w,(gt,t),h,l

    if type(testset) == tuple:
        if type(testset[0]) == str: # gold_path, pred_path
            testset = read_test(*testset)
        else: # tagger, dataset
            testset = process_pos_tagger(*testset, test=True)
    # otherwise, assume that testset is an iterable
    return testset

ACC, CONFUSION, STATS = range(3)

def eval_pos_tagger(testset, ignore_tags=[], by_type=None, dump_errors=None, listener=Listener, out=ACC):

    ignore_tags += ["<start>","ROOT"]
    testset = eval_read(testset)

    from functools import partial
    localzip = eval_localzip
    keep = partial(eval_keep, ignore_tags=ignore_tags)
    category = eval_category

    def config(tok): # _x, _y, y
        _, w, t = tok
        return w, t[-1], t[0]

    routine = eval_routine if out == ACC else matrix_routine if out == CONFUSION else stats_routine if out == STATS else None
    return eval_routine(listener.iter(testset), localzip, config, keep, category, eval_type=by_type, dump_errors=dump_errors)


def eval_localzip(ex):
    words, tags = ex[:2]
    return ((i,)+tok for i,tok in enumerate(zip(words, zip(*tags))))

def eval_keep(tok, ex=None, ignore_tags=[]):
    i,w,t = tok
    if t[0] in ignore_tags: return False
    return True

#
def eval_category(tok, ex, by_type):
    import math
    i,w,t = tok
    if type(by_type) == tuple:
        return tuple(category(tok, ex, sub_type) for sub_type in by_type)
    if type(by_type) == list:
        return "".join(str(category(tok, ex, sub_type)) for sub_type in by_type)
    if by_type == WORD:
        return w
    if by_type == POS:
        return t[0]
    return by_type

def strategies(tagger, strategy, itn, n_epoch):
    exploration_rate = 1

    if strategy==1:
        exploration_rate = 2*itn/n_epoch if itn < n_epoch / 2 else 1

    if strategy==2:
        exploration_rate = 2*itn/n_epoch if itn < n_epoch / 2 else 1
        tagger.lexicalization_level = int(itn/2)

    if strategy==3:
        tagger.lexicalization_dropout = True

    if strategy==4:
        pass

    return {"exploration_rate": exploration_rate}
