from .AbstractTransitionSystem import AbstractTransitionSystem

from dparser.BeamStackBufferState import BeamStackBufferState
from dparser.AbstractParser import AbstractParser
from dparser.AbstractBeamParser import AbstractBeamParser
from dparser.AbstractLocalParser import AbstractLocalParser

from dparser.BeamStackBufferState import LEFTMOST_CHILD, LEFT_CHILDREN, ID, RIGHT_CHILDREN, RIGHTMOST_CHILD

from dparser.AbstractParser import SHIFT, RIGHT, LEFT, REDUCE
MOVES = (SHIFT, RIGHT, LEFT, REDUCE)

# not compatible yet with root_first=True
nomultiroot = False

def nth(it, n, default=None):
    from itertools import islice
    return next(islice(it, n, None), default)
def len_gt(it, n): return n < 0 or nth(it, n) is not None

class Parser(AbstractParser):
    def __init__(self, features, *a, **kw):
        super().__init__(features, MOVES, *a, **kw)

class TransitionSystem(AbstractTransitionSystem):
    constant_length = True

################################################################################
##                                 LEGAL MOVES                                ##
################################################################################

    @staticmethod
    def _valid_moves(root_first, has_s0h, has_buffer, end_buffer, has_stack, nohead_stack):
        from itertools import chain
        arcdecomposable = not (root_first or nomultiroot)
        has_extended_stack = lambda: root_first or has_stack()
        if has_buffer():
            # a SHIFT calls for a LEFT
            if arcdecomposable or not (end_buffer() and has_extended_stack()):
                yield SHIFT
            if root_first or has_stack():
                stack_components = chain(nohead_stack(), [0] if root_first else [])
                # do not use last LEFT candidate
                if arcdecomposable or not (end_buffer() and len_gt(stack_components, 1)):
                    yield RIGHT
        if has_stack():
            if has_s0h():
                yield REDUCE
            else:
                yield LEFT

################################################################################
##                                CORRECT MOVES                               ##
################################################################################

    def _get_cut_edges(self, move, n0, s0, _stack, _nohead_stack, ref, root_first, _end_buffer, **kwargs):
        head_in_extended_stack, head_in_buffer, children_in_nohead_stack, children_in_buffer = self._auxcost_action(root_first, n0, _stack, _nohead_stack, *ref[:2])
        later_sacrifice, engaged_later_sacrifice = self._auxcost_arcdecomposability(n0, _stack, _nohead_stack, ref, root_first, _end_buffer)
        n0_leftmost, s0_leftmost, n0_rightmost, _, left_headed, right_headed = self._auxcost_fuzzy(_stack(), root_first, _end_buffer, ref[3])

        if move == SHIFT:
            yield from head_in_extended_stack(n0)
            yield from children_in_nohead_stack(n0)
            yield from later_sacrifice(0)
            if left_headed(n0) and not n0_leftmost(): yield 1
        if move == LEFT:
            yield from head_in_buffer(s0, 1)
            yield from children_in_buffer(s0)
            yield from engaged_later_sacrifice(start=1)
            if left_headed(n0) and s0_leftmost(): yield 1
        if move == RIGHT:
            yield from head_in_buffer(n0, 1)
            yield from head_in_extended_stack(n0, 1)
            yield from children_in_nohead_stack(n0)
            yield from later_sacrifice(1)
            if right_headed(n0) and not n0_rightmost(): yield 1
        if move == REDUCE:
            yield from children_in_buffer(s0)
            yield from engaged_later_sacrifice(start=1)
            if left_headed(n0) and s0_leftmost(): yield 1

################################################################################
##                                AUXILIARY COST                              ##
################################################################################

    @classmethod
    def _auxcost_action(cls, root_first, n0, _stack, _nohead_stack, head, children):
        def _extended_stack():
            yield from _stack()
            if root_first: yield 0
        return super()._auxcost_action(n0, _extended_stack, _nohead_stack, head, children)

    @staticmethod
    def _auxcost_arcdecomposability(n0, _stack, _nohead_stack, ref, root_first, _end_buffer):
        arcdecomposable = not (root_first or nomultiroot)
        _, children, (_, farright), _ = ref[:4]
        def needs_whole_buffer(si):
            ni = max(children[si]+[si])
            if ni < n0: return False
            return _end_buffer(farright[ni])
        def stack_components():
            from itertools import chain
            return chain(_nohead_stack(), [0] if root_first else [])
        def later_sacrifice(n):
            if not arcdecomposable and needs_whole_buffer(n0) and len_gt(stack_components(), n):
                return (yield 1)
            yield from engaged_later_sacrifice()
        def engaged_later_sacrifice(start=0):
            first_unbound = None
            if not arcdecomposable:
                first_unbound = nth(reversed(list(stack_components())), 1)
            for i, s in enumerate(_stack()):
                if first_unbound is None or s < first_unbound: return
                if i < start: continue
                if needs_whole_buffer(s): return (yield 1)
        return later_sacrifice, engaged_later_sacrifice

    @staticmethod
    def _auxcost_fuzzy(_stack, root_first, _end_buffer, fuzzy):
        def s_onleft(i): return not (root_first or len_gt(_stack(), i+1))
        def n_onright(ni): return root_first and _end_buffer(ni)
        def left_headed(c): return fuzzy[c:] and fuzzy[c] == -1
        def right_headed(c): return fuzzy[c] == +1
        return lambda:s_onleft(-1), lambda:s_onleft(0), lambda:n_onright(n0), lambda:n_onright(n0-1), left_headed, right_headed

################################################################################
##                                MOVE SEMANTICS                              ##
################################################################################

class BeamParse(TransitionSystem, BeamStackBufferState):
    def _valid_moves_elements(self):
        root_first = self.sentence[-2]
        has_s0h = lambda: self.s0h() is not None
        has_buffer = self.has_buffer
        end_buffer = self.end_buffer
        has_stack = self.has_stack
        nohead_stack = self.nohead_stack
        return root_first, has_s0h, has_buffer, end_buffer, has_stack, nohead_stack

    def transition(self, move, score=0, label=None):
        if move == SHIFT:
            n0 = (None,None,0,self.n0[ID]+1,0,None,None)
            s0 = self.n0
            tail = self.push()
        elif move == RIGHT:
            if label is not None: label = self.n0[ID], label
            child = label if label else self.n0[ID]
            n0 = (None,None,0,self.n0[ID]+1,0,None,None)
            s0 = self.n0
            s1 = self.s0[:ID+1] + (self.s0[RIGHT_CHILDREN]+1, self.s0[RIGHTMOST_CHILD], child)
            tail = self.__class__(s0, s1, self.stack_tail, (self,move,score)).push()
        elif move == LEFT:
            if label is not None: label = self.s0[ID], label
            child = label if label else self.s0[ID]
            n0 = (child, self.n0[LEFTMOST_CHILD], self.n0[LEFT_CHILDREN]+1) + self.n0[ID:]
            s0 = self.stack_tail.s0
            tail = self.pop()
        elif move == REDUCE:
            n0 = self.n0
            s0 = self.stack_tail.s0
            tail = self.pop()
        assert move in MOVES
        new = self.__class__(n0, s0, tail, (self,move,score))
        if type(label) == tuple: new.last_label = label
        return new

    @property
    def heads(self):
        heads = [None for _ in self.sentence[0]]
        labels = [None for _ in self.sentence[0]]
        labeled = False
        for parse in self.predecessors:
            if parse.last_move == RIGHT:
                heads[parse.previous.n0[ID]] = parse.previous.s0[ID]
                labeled |= parse.set_label(labels)
            elif parse.last_move == LEFT:
                heads[parse.previous.s0[ID]] = parse.previous.n0[ID]
                labeled |= parse.set_label(labels)
        root_first = self.sentence[-2]
        if not root_first:
            heads[-1] = 0
        if labeled: return heads, labels
        return heads

    @property
    def s0_ancestors(self):
        for tail in self._stack_tails:
            s1 = tail.stack_tail.s0
            if type(s1[RIGHTMOST_CHILD]) == tuple:
                if s1[RIGHTMOST_CHILD][0] != tail.s0[ID]:
                    break
            else:
                if s1[RIGHTMOST_CHILD] != tail.s0[ID]:
                    break
            yield s1[ID]

    @property
    def final(self):
        return not self.has_stack() and not self.has_buffer()

    @property
    def length_final(self):
        root_first, n = self.sentence[-2:]
        return 2 * n - (0 if root_first else 1)

class BeamParser(Parser, AbstractBeamParser):
    def _defaultBeamParse(self):
        return BeamParse()

    def _initBeamParse(self, sample, parse=None):
        parse = super()._initBeamParse(sample, parse)
        if not parse.depth and not self.root_first:
            parse = parse + SHIFT
            parse.last_move = None
            parse.previous = None
        return parse

class LocalParser(TransitionSystem, Parser, AbstractLocalParser):
    def _valid_moves_elements(self, n0, n, stack, heads):
        root_first = self.root_first
        has_s0h = lambda: stack and heads[stack[-1]] is not None
        has_buffer = lambda: n0 <= n
        end_buffer = lambda: n0 == n
        has_stack = lambda: stack and True
        nohead_stack = lambda: (s for s in reversed(stack) if heads[s] is None)
        return root_first, has_s0h, has_buffer, end_buffer, has_stack, nohead_stack

    def transition(self, move, n0, stack, parse):
        if move == SHIFT:
            stack.append(n0)
            return n0 + 1
        elif move == RIGHT:
            parse.add(stack[-1], n0)
            stack.append(n0)
            return n0 + 1
        elif move == LEFT:
            parse.add(n0, stack.pop())
            return n0
        elif move == REDUCE:
            stack.pop()
            return n0
        assert move in MOVES

    def final(self, n0, n, stack):
        return not stack and n0 > n
