from .AbstractTransitionSystem import AbstractTransitionSystem

from dparser.BeamStackBufferState import BeamStackBufferState
from dparser.AbstractParser import AbstractParser
from dparser.AbstractBeamParser import AbstractBeamParser
from dparser.AbstractLocalParser import AbstractLocalParser

from dparser.BeamStackBufferState import LEFTMOST_CHILD, LEFT_CHILDREN, ID, RIGHT_CHILDREN, RIGHTMOST_CHILD

from dparser.AbstractParser import SHIFT, RIGHT, LEFT
MOVES = (SHIFT, RIGHT, LEFT)

# unchecked
nomultiroot = False

class Parser(AbstractParser):
    def __init__(self, features, *a, **kw):
        super().__init__(features, MOVES, *a, **kw)

class TransitionSystem(AbstractTransitionSystem):
    constant_length = True

################################################################################
##                                 LEGAL MOVES                                ##
################################################################################

    @staticmethod
    def _valid_moves(root_first, has_token_s1, has_buffer, has_stack):
        if has_buffer():
            yield SHIFT
        if has_stack():
            if root_first or has_token_s1():
                yield RIGHT
            if not (root_first or nomultiroot and has_token_s1()) or has_buffer():
                yield LEFT

################################################################################
##                                CORRECT MOVES                               ##
################################################################################

    def _get_cut_edges(self, move, n0, s0, _stack, _nohead_stack, ref, root_first, _end_buffer, **kwargs):
        head_in_extended_stack, head_in_buffer, children_in_nohead_stack, children_in_buffer = self._auxcost_action(root_first, n0, _stack, _nohead_stack, *ref[:2])

        if move == SHIFT:
            yield from head_in_extended_stack(n0, 1)
            yield from children_in_nohead_stack(n0)
        if move == LEFT:
            yield from head_in_buffer(s0, 1)
            yield from head_in_extended_stack(s0, 1, 1)
            yield from children_in_buffer(s0)
        if move == RIGHT:
            yield from head_in_buffer(s0)
            yield from children_in_buffer(s0)

################################################################################
##                                AUXILIARY COST                              ##
################################################################################

    @classmethod
    def _auxcost_action(cls, root_first, n0, _stack, _nohead_stack, head, children):
        def _extended_stack():
            yield from _stack()
            if root_first: yield 0
        return super()._auxcost_action(n0, _extended_stack, _nohead_stack, head, children)

################################################################################
##                                MOVE SEMANTICS                              ##
################################################################################

class BeamParse(TransitionSystem, BeamStackBufferState):
    def _valid_moves_elements(self):
        root_first = self.sentence[-2]
        has_token_s1 = lambda: self.s(1)
        has_buffer = self.has_buffer
        has_stack = self.has_stack
        return root_first, has_token_s1, has_buffer, has_stack

    def transition(self, move, score=0, label=None):
        if move == SHIFT:
            n0 = (None,None,0,self.n0[ID]+1,0,None,None)
            s0 = self.n0
            tail = self.push()
        elif move == RIGHT:
            _s1 = self.stack_tail.s0
            n0 = self.n0
            s0 = _s1[:ID+1] + (_s1[RIGHT_CHILDREN]+1, _s1[RIGHTMOST_CHILD], self.s0[ID])
            tail = self.pop()
            if label is not None: label = self.s0[ID], label
        elif move == LEFT:
            n0 = (self.s0[ID], self.n0[LEFTMOST_CHILD], self.n0[LEFT_CHILDREN]+1) + self.n0[ID:]
            s0 = self.stack_tail.s0
            tail = self.pop()
            if label is not None: label = self.s0[ID], label
        assert move in MOVES
        new = self.__class__(n0, s0, tail, (self,move,score))
        if type(label) == tuple: new.last_label = label
        return new

    @property
    def heads(self):
        heads = [None for _ in self.sentence[0]]
        labels = [None for _ in self.sentence[0]]
        labeled = False
        for parse in self.predecessors:
            if parse.last_move == RIGHT:
                heads[parse.previous.s0[ID]] = parse.s0[ID]
                labeled |= parse.set_label(labels)
            elif parse.last_move == LEFT:
                heads[parse.previous.s0[ID]] = parse.previous.n0[ID]
                labeled |= parse.set_label(labels)
        root_first = self.sentence[-2]
        if not root_first:
            heads[-1] = 0
        if labeled: return heads, labels
        return heads

    @property
    def s0_ancestors(self):
        return
        yield

    @property
    def final(self):
        return not self.has_stack() and not self.has_buffer()

    @property
    def length_final(self):
        root_first, n = self.sentence[-2:]
        return 2 * n - (0 if root_first else 1)

class BeamParser(Parser, AbstractBeamParser):
    def _defaultBeamParse(self):
        return BeamParse()

    def _initBeamParse(self, sample, parse=None):
        parse = super()._initBeamParse(sample, parse)
        if not parse.depth and not self.root_first:
            parse = parse + SHIFT
            parse.last_move = None
            parse.previous = None
        return parse

class LocalParser(TransitionSystem, Parser, AbstractLocalParser):
    def _valid_moves_elements(self, n0, n, stack, heads):
        root_first = self.root_first
        has_token_s1 = lambda: len(stack) >= 2
        has_buffer = lambda: n0 <= n
        has_stack = lambda: stack and True
        return root_first, has_token_s1, has_buffer, has_stack

    def transition(self, move, n0, stack, parse):
        if move == SHIFT:
            stack.append(n0)
            return n0 + 1
        elif move == RIGHT:
            s1 = stack[-2] if len(stack) >= 2 else 0
            parse.add(s1, stack.pop())
            return n0
        elif move == LEFT:
            parse.add(n0, stack.pop())
            return n0
        assert move in MOVES

    def final(self, n0, n, stack):
        return not stack and n0 > n
