from .AbstractTransitionSystem import AbstractTransitionSystem

from dparser.BeamStackBufferState import BeamStackBufferState
from dparser.AbstractParser import AbstractParser
from dparser.AbstractBeamParser import AbstractBeamParser
from dparser.AbstractLocalParser import AbstractLocalParser

from dparser.BeamStackBufferState import LEFTMOST_CHILD, LEFT_CHILDREN, ID, RIGHT_CHILDREN, RIGHTMOST_CHILD

from dparser.AbstractParser import SHIFT, RIGHT, LEFT
MOVES = (SHIFT, RIGHT, LEFT)

# unchecked
nomultiroot = False

class Parser(AbstractParser):

    def __init__(self, features, *a, **kw):
        super().__init__(features, MOVES, *a, **kw)

class TransitionSystem(AbstractTransitionSystem):
    constant_length = True

################################################################################
##                                 LEGAL MOVES                                ##
################################################################################

    @staticmethod
    def _valid_moves(root_first, has_token_s1, has_extended_buffer, has_stack):
        if has_extended_buffer():
            yield SHIFT
        if has_stack():
            has_token_s1 = has_token_s1()
            if root_first or has_token_s1 and has_extended_buffer():
                yield RIGHT
            if has_token_s1 or not root_first and not has_extended_buffer():
                yield LEFT

################################################################################
##                                CORRECT MOVES                               ##
################################################################################

    def _get_cut_edges(self, move, n0, s0, _stack, _nohead_stack, ref, root_first, _end_buffer, **kwargs):
        head_in_extended_stack, head_in_buffer, children_in_stack, children_in_buffer = self._auxcost_action(root_first, n0, _stack, _nohead_stack, *ref[:2])

        head, children, (farleft, farright) = ref[:3]
        def si_candidate_ancestor_for_ni(si=0, ni=n0):
            while type(head[ni]) == int and head[ni] >= n0:
                ni = head[ni]
            return next(head_in_extended_stack(ni, si+1), None) is None

        def _right_stack(left_stack, n0):
            yield left_stack[0]
            ni = n0
            inserted = set()
            left_stack = set(left_stack)
            while not (root_first and _end_buffer(ni-1) or _end_buffer(ni-2)):
                if (head[ni] is None or head[ni] < n0) or any(c in left_stack or c in inserted for c in children[ni]):
                    inserted.add(ni)
                    yield ni
                ni += 1

        def C(h,c): return int(head[c] != h and head[c] is not None)

        def loss(_stack, n0):
            window = []
            left_stack = list(_stack()) + ([0] if root_first else [])
            right_stack = list(_right_stack(left_stack, n0))
            for j,nj in enumerate(right_stack):
                for i,si in enumerate(left_stack):
                    expand_left = window[-1] if i else {} # t[i-1,j]
                    expand_right = window.pop(0) if j else {} # t[i,j-1]
                    new = {} # t[i,j]
                    window.append(new)
                    candidates = [C(si,h)+loss for h,loss in expand_left.items()]
                    if j: candidates += [expand_right[si] + C(si,nj)]
                    new[si] = min(candidates) if candidates else 0
                    for h in left_stack[:i] + right_stack[:j]:
                        candidates = []
                        if i: candidates += [C(h,si) + expand_left[h]]
                        if j: candidates += [expand_right[h] + C(h,nj)]
                        new[h] = min(candidates)
                    candidates = [expand_right[h]+C(nj,h) for h, loss in expand_right.items()]
                    if i: candidates += [C(nj,si) + expand_left[nj]]
                    new[nj] = min(candidates) if candidates else 0
            return window[-1][0 if root_first else nj]

        before = loss(_stack, n0)

        from itertools import chain, islice
        if move == SHIFT:
            after = loss(lambda: chain([n0], _stack()), n0 + 1)
            for _ in range(before, after):
                yield 1
        if move == LEFT:
            stack = chain(_stack(), [0] if root_first else [])
            _, s1 = next(stack), next(stack)
            after = loss(lambda: chain((s for i,s in enumerate(_stack()) if i != 1)), n0) + C(s0,s1)
            for _ in range(before, after):
                yield 1
        if move == RIGHT:
            stack = chain(_stack(), [0] if root_first else [])
            _, s1 = next(stack), next(stack)
            after = loss(lambda: chain((s for i,s in enumerate(_stack()) if i != 0)), n0) + C(s1,s0)
            for _ in range(before, after):
                yield 1

################################################################################
##                                AUXILIARY COST                              ##
################################################################################

    @classmethod
    def _auxcost_action(cls, root_first, n0, _stack, _nohead_stack, head, children):
        def _extended_stack():
            yield from _stack()
            if root_first: yield 0
        return super()._auxcost_action(n0, _extended_stack, _nohead_stack, head, children)

################################################################################
##                                MOVE SEMANTICS                              ##
################################################################################

class BeamParse(TransitionSystem, BeamStackBufferState):
    def _valid_moves_elements(self):
        root_first = self.sentence[-2]
        has_token_s1 = lambda: self.s(1)
        has_extended_buffer = self.has_extended_buffer
        has_stack = self.has_stack
        return root_first, has_token_s1, has_extended_buffer, has_stack

    def transition(self, move, score=0, label=None):
        if move == SHIFT:
            n0 = (None,None,0,self.n0[ID]+1,0,None,None)
            if not self.has_buffer():
                s0 = self.s0
                tail = self.stack_tail
            else:
                s0 = self.n0
                tail = self
        elif move == RIGHT:
            _s1 = self.stack_tail.s0
            n0 = self.n0
            s0 = _s1[:ID+1] + (_s1[RIGHT_CHILDREN]+1, _s1[RIGHTMOST_CHILD], self.s0[ID])
            tail = self.pop()
            if label is not None: label = self.s0[ID], label
        elif move == LEFT:
            n0 = self.n0
            if self.uses_extended_buffer():
                s0 = self.stack_tail.s0
                if label is not None: label = self.s0[ID], label
            else:
                _s1 = self.stack_tail.s0
                if label is not None: label = _s1[ID], label
                s0 = (_s1[ID], self.s0[LEFTMOST_CHILD], self.s0[LEFT_CHILDREN]+1) + self.s0[ID:]
            tail = self.pop()
        assert move in MOVES
        new = self.__class__(n0, s0, tail, (self,move,score))
        if type(label) == tuple: new.last_label = label
        return new

    @property
    def heads(self):
        heads = [None for _ in self.sentence[0]]
        labels = [None for _ in self.sentence[0]]
        labeled = False
        for parse in self.predecessors:
            if parse.last_move == RIGHT:
                heads[parse.previous.s0[ID]] = parse.s0[ID]
                labeled |= parse.set_label(labels)
            elif parse.last_move == LEFT:
                if parse.uses_extended_buffer():
                    heads[parse.previous.s0[ID]] = parse.n0[ID] - 1
                else:
                    heads[parse.previous.stack_tail.s0[ID]] = parse.s0[ID]
                labeled |= parse.set_label(labels)
        root_first = self.sentence[-2]
        if not root_first:
            heads[-1] = 0
        if labeled: return heads, labels
        return heads

    @property
    def s0_ancestors(self):
        return
        yield

    def uses_extended_buffer(self):
        return self.end_buffer(self.n0[ID]-2)

    def has_extended_buffer(self):
        root_first = self.sentence[-2]
        return self.has_buffer() or (not root_first and not self.uses_extended_buffer())

    @property
    def final(self):
        return not self.has_stack() and not self.has_extended_buffer()

    @property
    def length_final(self):
        return 2 * self.sentence[-1]

class BeamParser(Parser, AbstractBeamParser):
    def _defaultBeamParse(self):
        return BeamParse()

    def _initBeamParse(self, sample, parse=None):
        parse = super()._initBeamParse(sample, parse)
        #
        if not parse.depth and not self.root_first:
            parse = parse + SHIFT
            parse.last_move = None
            parse.previous = None
        return parse

class LocalParser(TransitionSystem, Parser, AbstractLocalParser):
    def _valid_moves_elements(self, n0, n, stack, heads):
        root_first = self.root_first
        has_token_s1 = lambda: len(stack) >= 2
        has_extended_buffer = lambda: n0 <= n or (not root_first and n0 == n + 1)
        has_stack = lambda: stack and True
        return root_first, has_token_s1, has_extended_buffer, has_stack

    def transition(self, move, n0, stack, parse):
        if move == SHIFT:
            stack.append(n0)
            return n0 + 1
        elif move == RIGHT:
            s1 = stack[-2] if len(stack) >= 2 else 0
            parse.add(s1, stack.pop())
            return n0
        elif move == LEFT:
            parse.add(stack[-1], stack.pop(-2))
            return n0
        assert move in MOVES

    def final(self, n0, n, stack):
        return (self.root_first and not stack or len(stack) < 2) and n0 > n
